from typing import Callable

import pytest
import torch

from elasticai.creator.arithmetic import FxpConverter, FxpParams
from elasticai.creator.nn.fixed_point.conv1d.design import Conv1dDesign
from elasticai.creator.nn.fixed_point.conv1d.testbench import Conv1dTestbench
from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.design.ports import Port


class DummyConv1d:
    def __init__(self, fxp_params: FxpParams, in_channels: int, out_channels: int):
        self.name: str = "conv1d"
        self.kernel_size: int = 1
        self.input_signal_length = 1
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.port: Port = create_port(
            y_width=fxp_params.total_bits,
            x_width=fxp_params.total_bits,
            x_count=1,
            y_count=2,
        )


def parameters_for_reported_content_parsing(fxp_params, input_expected_pairs):
    def add_expected_prefix_to_pairs(pairs):
        _converter_for_batch = FxpConverter(
            FxpParams(total_bits=8, frac_bits=0, signed=True)
        )  # max for 255 lines of inputs
        pairs_with_prefix = list()
        for i, (pairs_text, pairs_number) in enumerate(pairs):
            pairs_with_prefix.append(list())
            pairs_with_prefix[i].append(list())
            pairs_with_prefix[i].append(pairs_number)
            for batch_number, batch_channel_text in enumerate(pairs_text):
                for out_channel_text in batch_channel_text:
                    for value_text in out_channel_text:
                        pairs_with_prefix[i][0].append(
                            f"result: {_converter_for_batch.integer_to_binary_string_vhdl(batch_number)},"
                            f" {value_text}"
                        )
        return pairs_with_prefix

    pairs_with_prefix = [
        (fxp_params, a, b)
        for a, b in add_expected_prefix_to_pairs(input_expected_pairs)
    ]
    return pairs_with_prefix


@pytest.fixture
def create_uut() -> Callable[[FxpParams, int, int], Conv1dDesign]:
    def create(
        fxp_params: FxpParams, in_channels: int, out_channels: int
    ) -> Conv1dDesign:
        return DummyConv1d(
            fxp_params, in_channels=in_channels, out_channels=out_channels
        )

    return create


@pytest.mark.parametrize(
    "fxp_params, reported, expected",
    (
        parameters_for_reported_content_parsing(
            fxp_params=FxpParams(total_bits=3, frac_bits=0),
            input_expected_pairs=[
                ([[["010"]]], [[[2.0]]]),
                ([[["111"]]], [[[-1.0]]]),
                ([[["001", "010"]]], [[[1.0, 2.0]]]),
                ([[["111", "001"]]], [[[-1.0, 1.0]]]),
            ],
        )
        + parameters_for_reported_content_parsing(
            fxp_params=FxpParams(total_bits=4, frac_bits=1),
            input_expected_pairs=[
                ([[["0001", "1111"]]], [[[0.5, -0.5]]]),
                (
                    [[["0001", "0011"]], [["1000", "1111"]]],
                    [[[0.5, 1.5]], [[-4.0, -0.5]]],
                ),
            ],
        )
    ),
)
def test_parse_reported_content_one_out_channel(
    fxp_params, reported, expected, create_uut
):
    in_channels = None
    out_channels = 1
    bench = Conv1dTestbench(
        name="conv1d_testbench",
        fxp_params=fxp_params,
        uut=create_uut(fxp_params, in_channels, out_channels),
    )
    result = bench.parse_reported_content(reported)
    assert result == expected


@pytest.mark.parametrize(
    "fxp_params, reported, expected",
    (
        parameters_for_reported_content_parsing(
            fxp_params=FxpParams(total_bits=3, frac_bits=0),
            input_expected_pairs=[
                ([[["010"], ["010"]]], [[[2.0], [2.0]]]),
                ([[["001", "010"], ["001", "010"]]], [[[1.0, 2.0], [1.0, 2.0]]]),
                ([[["111", "001"], ["111", "001"]]], [[[-1.0, 1.0], [-1.0, 1.0]]]),
            ],
        )
        + parameters_for_reported_content_parsing(
            fxp_params=FxpParams(total_bits=4, frac_bits=1),
            input_expected_pairs=[
                ([[["0001", "1111"], ["0001", "1111"]]], [[[0.5, -0.5], [0.5, -0.5]]]),
                (
                    [
                        [["0001", "0011"], ["0001", "0011"]],
                        [["1000", "1111"], ["1000", "1111"]],
                    ],
                    [[[0.5, 1.5], [0.5, 1.5]], [[-4.0, -0.5], [-4.0, -0.5]]],
                ),
            ],
        )
    ),
)
def test_parse_reported_content_two_out_channel(
    fxp_params, reported, expected, create_uut
):
    in_channels = None
    out_channels = 2
    bench = Conv1dTestbench(
        name="conv1d_testbench",
        fxp_params=fxp_params,
        uut=create_uut(fxp_params, in_channels, out_channels),
    )
    result = bench.parse_reported_content(reported)
    assert result == expected


def test_input_preparation_with_one_in_channel(create_uut):
    fxp_params = FxpParams(total_bits=3, frac_bits=0)
    in_channels = 1
    out_channels = None
    bench = Conv1dTestbench(
        name="bench_name",
        fxp_params=fxp_params,
        uut=create_uut(fxp_params, in_channels, out_channels),
    )
    input = torch.Tensor([[[1.0, 1.0]]])
    expected = [{"x_0_0": "001", "x_0_1": "001"}]
    result = bench.prepare_inputs(input.tolist())
    assert result == expected


def test_input_preparation_with_two_in_channel(create_uut):
    fxp_params = FxpParams(total_bits=3, frac_bits=0)
    in_channels = 1
    out_channels = None
    bench = Conv1dTestbench(
        name="bench_name",
        fxp_params=fxp_params,
        uut=create_uut(fxp_params, in_channels, out_channels),
    )
    input = torch.Tensor([[[1.0, 1.0], [1.0, 2.0]]])
    expected = [{"x_0_0": "001", "x_0_1": "001", "x_1_0": "001", "x_1_1": "010"}]
    result = bench.prepare_inputs(input.tolist())
    assert result == expected



from dataclasses import dataclass
from math import ceil
from typing import cast

from elasticai.creator.arithmetic.fxp_arithmetic import FxpArithmetic
from elasticai.creator.arithmetic.fxp_params import (
    ConvertableToFixedPointValues,
    FxpParams,
    T,
)


@dataclass
class FxpConverter:
    def __init__(self, fxp_params: FxpParams):
        self._config = fxp_params

    @property
    def total_bits(self) -> int:
        return self._config.total_bits

    @property
    def frac_bits(self) -> int:
        return self._config.frac_bits

    def _round_to_integer(self, number: float | T) -> int | T:
        return FxpArithmetic(self._config).round_to_integer(number)

    def _integer_to_twos(self, number):
        if isinstance(number, ConvertableToFixedPointValues):
            return int(self._convert_integer_to_twos(cast(T, number)))
        else:
            return self._convert_integer_to_twos(number)

    def _convert_integer_to_twos(self, number):
        if self._config.integer_out_of_bounds(number):
            raise ValueError(
                f"Value '{number}' cannot be represented with {self._config.total_bits} bits."
            )
        return ((1 << self._config.total_bits) + number) if number < 0 else number

    def integer_to_binary_string_vhdl(self, number: int) -> str:
        twos = self._integer_to_twos(number)
        return f'"{twos:0{self._config.total_bits}b}"'

    def integer_to_hex_string_vhdl(self, number: int) -> str:
        twos = self._integer_to_twos(number)
        return f'X"{twos:0{ceil(self._config.total_bits / 4)}X}"'

    def integer_to_binary_string_verilog(self, number: int) -> str:
        twos = self._integer_to_twos(number)
        return f"{self._config.total_bits}'b{twos:0{self._config.total_bits}b}"

    def integer_to_decimal_string_verilog(self, number: int) -> str:
        twos = self._integer_to_twos(number)
        return f"{self._config.total_bits}'d{twos}"

    def integer_to_hex_string_verilog(self, number: int) -> str:
        twos = self._integer_to_twos(number)
        return (
            f"{self._config.total_bits}'h{twos:0{ceil(self._config.total_bits / 4)}X}"
        )

    def rational_to_binary_string_vhdl(self, number: float) -> str:
        fxp = self._round_to_integer(number)
        return self.integer_to_binary_string_vhdl(fxp)

    def rational_to_hex_string_vhdl(self, number: float) -> str:
        fxp = self._round_to_integer(number)
        return self.integer_to_hex_string_vhdl(fxp)

    def rational_to_binary_string_verilog(self, number: float) -> str:
        fxp = self._round_to_integer(number)
        return self.integer_to_binary_string_verilog(fxp)

    def rational_to_decimal_string_verilog(self, number: float) -> str:
        fxp = self._round_to_integer(number)
        return self.integer_to_decimal_string_verilog(fxp)

    def rational_to_hex_string_verilog(self, number: float) -> str:
        fxp = self._round_to_integer(number)
        return self.integer_to_hex_string_verilog(fxp)

    def binary_to_integer(self, binary: str) -> int:
        format_binary = binary.replace('"', "").replace(" ", "").split("b")[-1]
        return int(format_binary, 2) - (
            (1 << self._config.total_bits)
            if self._config.signed and format_binary[0] == "1"
            else 0
        )

    def binary_to_rational(self, binary: str) -> float:
        int_val = self.binary_to_integer(binary)
        return int_val * self._config.minimum_step_as_rational

    def decimal_to_integer(self, binary: str) -> int:
        format_binary = binary.split("d")[-1]
        return int(format_binary) - (
            (1 << self._config.total_bits)
            if self._config.signed and format_binary[0] == "1"
            else 0
        )

    def decimal_to_rational(self, binary: str) -> float:
        int_val = self.decimal_to_integer(binary)
        return int_val * self._config.minimum_step_as_rational

    def hex_to_integer(self, binary: str) -> int:
        format_binary = binary.replace('"', "").replace(" ", "").split("X")[-1]
        return int(format_binary, 16) - (
            (1 << self._config.total_bits)
            if self._config.signed and format_binary[0] == "1"
            else 0
        )

    def hex_to_rational(self, binary: str) -> float:
        int_val = self.hex_to_integer(binary)
        return int_val * self._config.minimum_step_as_rational



import pytest
import torch

from elasticai.creator.base_modules.linear import Linear
from elasticai.creator.nn.binary.math_operations import MathOperations

operations = MathOperations()


class TestBinaryOperations:
    def test_quantization_returns_1_for_0_8(self) -> None:
        x = torch.tensor([0.8])
        x = operations.quantize(x)
        x = x.tolist()
        assert x == [1.0]

    def test_quantization_returns_1_for_2_3(self) -> None:
        x = torch.tensor([2.3])
        x = operations.quantize(x)
        x = x.tolist()
        assert x == [1.0]

    def test_quantization_returns_minus_1_for_minus_0_7(self) -> None:
        x = torch.tensor([-0.7])
        x = operations.quantize(x)
        x = x.tolist()
        assert x == [-1.0]

    def test_quantization_returns_minus_1_for_minus_3(self) -> None:
        x = torch.tensor([-3])
        x = operations.quantize(x)
        x = x.tolist()
        assert x == [-1.0]

    def test_quantization_produces_gradient_1_for_minus_0_3(self) -> None:
        x = torch.tensor([-0.3], requires_grad=True)
        y = operations.quantize(x)
        y.backward()
        g = x.grad
        assert g.tolist() == [1.0]

    def test_quantization_produces_gradient_0_for_3(self) -> None:
        x = torch.tensor([3.0], requires_grad=True)
        y = operations.quantize(x)
        y.backward()
        g = x.grad
        assert g.tolist() == [0.0]

    @pytest.mark.parametrize(
        argnames=("a", "b", "result"),
        argvalues=(
            (-1, 1, 1),
            (-1, -1, -1),
            (1, 1, 1),
        ),
    )
    def test_addition(self, a, b, result) -> None:
        x = torch.tensor([float(a)])
        y = torch.tensor([float(b)])
        assert operations.add(x, y) == torch.tensor([float(result)])

    @pytest.fixture(params=((-1, -1, 1), (1, 1, -1), (-1, -1, -1), (1, 1, 1)))
    def operants(self, request):
        return request.param

    @pytest.fixture
    def operants_as_tensors(self, operants) -> torch.Tensor:
        return torch.tensor([float(x) for x in operants])

    def test_matmul(self) -> None:
        a = [[-1.0, -1.0], [1.0, -1]]
        b = [[1.0, 1.0], [-1.0, 1.0]]
        expected = [[1.0, -1.0], [1.0, 1.0]]
        actual = operations.matmul(torch.tensor(a), torch.tensor(b)).tolist()
        assert actual == expected

    def test_mul(self) -> None:
        a = [-1.0, -1.0, 1.0, 1.0]
        b = [1.0, -1.0, -1.0, 1.0]
        expected = [-1.0, 1.0, -1.0, 1.0]
        actual = operations.mul(torch.tensor(a), torch.tensor(b)).tolist()
        assert actual == expected


class TestBinaryOperationsWithLinearLayer:
    def test_binary_linear_layer(self) -> None:
        x = [-0.4, 0.1]
        weight = [[0.1, -0.8], [-1.2, 4.0]]
        expected = [-1.0, 1.0]
        linear = Linear(
            operations=MathOperations(),
            in_features=len(x),
            out_features=len(weight),
            bias=False,
        )
        with torch.no_grad():
            linear.weight = torch.nn.Parameter(torch.tensor(weight))

        actual = linear(torch.tensor(x)).tolist()
        assert actual == expected



from typing import cast

import torch

from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.nn.design_creator_module import DesignCreatorModule
from elasticai.creator.nn.fixed_point.math_operations import MathOperations
from elasticai.creator.vhdl.shared_designs.precomputed_scalar_function import (
    PrecomputedScalarFunction,
)

from .identity_step_function import IdentityStepFunction


class PrecomputedModule(DesignCreatorModule):
    _xoffset: float
    _lut_input: torch.nn.Buffer

    def __init__(
        self,
        base_module: torch.nn.Module,
        total_bits: int,
        frac_bits: int,
        num_steps: int,
        sampling_intervall: tuple[float, float],
    ) -> None:
        super().__init__()
        self._base_module = base_module
        self._params = FxpParams(
            total_bits=total_bits, frac_bits=frac_bits, signed=True
        )
        self._config = FxpArithmetic(self._params)
        self._operations = MathOperations(self._config)
        self._build_lut(in_range=sampling_intervall, lut_size=num_steps)

    def _build_lut(self, in_range: tuple[float, float], lut_size: int) -> None:
        range_neg = (
            self._config.minimum_as_rational
            if abs(in_range[0]) == float("inf")
            else in_range[0]
        )
        range_pos = (
            self._config.maximum_as_rational
            if abs(in_range[1]) == float("inf")
            else in_range[1]
        )
        lut_num_steps = (
            2**self._config.total_bits
            if lut_size > 2**self._config.total_bits
            else lut_size
        )

        self._lut_input = torch.nn.Buffer(
            self._operations.round(
                torch.linspace(start=range_neg, end=range_pos, steps=lut_num_steps)
            ),
            persistent=False,
        )
        lut_diff = (
            torch.abs(torch.diff(self._lut_input))
            / self._params.minimum_step_as_rational
            / 2
        )
        self._xoffset = (
            float((lut_diff.max() + lut_diff.min()) / 2)
            * self._params.minimum_step_as_rational
        )

    def get_lut_integer(self) -> tuple[list[int], list[int]]:
        return (
            list(map(self._config.cut_as_integer, self._lut_input.tolist())),
            [self._forward_nograd(val) for val in self._lut_input.tolist()],
        )

    def _stepped_inputs(self, x: torch.Tensor) -> torch.Tensor:
        return cast(torch.Tensor, IdentityStepFunction.apply(x, self._lut_input))

    def _forward_nograd(self, x: int) -> int:
        fxp_input = self._config.as_rational(x)
        if not isinstance(fxp_input, float):
            raise ValueError()

        with torch.no_grad():
            output = self.forward(torch.tensor(fxp_input).clone().detach())
        return self._config.round_to_integer(float(output.item()))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self._stepped_inputs(x)
        y = self._base_module(x - self._xoffset)
        return self._operations.round(y)

    def create_design(self, name: str) -> PrecomputedScalarFunction:
        q_input = self.get_lut_integer()[0]
        return PrecomputedScalarFunction(
            name=name,
            input_width=self._config.total_bits,
            output_width=self._config.total_bits,
            inputs=q_input,
            function=self._forward_nograd,
        )



from collections.abc import Callable, Iterator
from dataclasses import dataclass

import pytest

import elasticai.creator.plugin as pl
from elasticai.creator.ir import Lowerable, LoweringPass


@dataclass
class PluginSpec(pl.PluginSpec):
    generated: list[str]


@pytest.fixture
def plugin() -> PluginSpec:
    p = list(
        pl.read_plugin_dicts_from_package("tests.integration_tests.minimal_plugin")
    )[0]
    return pl.build_plugin_spec(p, PluginSpec)


def test_can_read_plugin(plugin) -> None:
    p = plugin
    assert (
        PluginSpec(
            name="minimal_plugin",
            target_platform="elastic-node-v5",
            target_runtime="vhdl",
            version="0.1",
            api_version="0.1",
            generated=[
                "convolution",
            ],
            package="tests.integration_tests.minimal_plugin",
        )
        == p
    )


class DummyLowerable(Lowerable):
    def __init__(self, type: str, more_data: list[str]):
        self._type = type
        self.more_data: list[str] = more_data

    @property
    def type(self) -> str:
        return self._type


@pytest.fixture
def make_lowerable(plugin) -> Callable[[list[str]], DummyLowerable]:
    def dummy(more_data: list[str]) -> DummyLowerable:
        return DummyLowerable(plugin.generated[0], more_data)

    return dummy


def test_can_load_entire_plugin(
    make_lowerable: Callable[[list[str]], DummyLowerable],
) -> None:
    lower: LoweringPass[DummyLowerable, str] = LoweringPass()

    def extract_symbols(p: PluginSpec) -> Iterator[pl.PluginSymbol]:
        if p.target_runtime == "vhdl":
            yield from pl.import_symbols(f"{p.package}.src", set(p.generated))

    loader = pl.PluginLoader(
        fetch=pl.SymbolFetcherBuilder(PluginSpec).add_fn(extract_symbols).build(),
        plugin_receiver=lower,
    )
    lowerable = make_lowerable(["some", "important", "information"])
    loader.load_from_package("tests.integration_tests.minimal_plugin")

    assert ("some_important_information",) == tuple(lower((lowerable,)))


def test_can_load_lowering_pass_plugin(
    make_lowerable: Callable[[list[str]], DummyLowerable],
) -> None:
    lower: LoweringPass[DummyLowerable, str] = LoweringPass()

    def fetch(p: PluginSpec) -> Iterator[pl.PluginSymbol]:
        if p.target_runtime == "lowering_pass_test":
            yield from pl.import_symbols(f"{p.package}.src", set(p.generated))

    loader = pl.PluginLoader(
        fetch=pl.SymbolFetcherBuilder(PluginSpec).add_fn(fetch).build(),
        plugin_receiver=lower,
    )
    lowerable = DummyLowerable(
        "lowering_pass_conv", ["some", "important", "information"]
    )
    loader.load_from_package("tests.integration_tests.minimal_plugin")
    print(lower.__dict__["_fns"].__dict__["_fns"])

    assert ("some_important_information",) == tuple(lower((lowerable,)))



import pytest
import torch

import elasticai.creator.nn.fixed_point as nn_fxp
from elasticai.creator.arithmetic import FxpParams
from elasticai.creator.nn import Sequential


class SimpleLinear(torch.nn.Module):
    def __init__(self, total_width: int, frac_width: int):
        super().__init__()
        self.model = Sequential(
            nn_fxp.Linear(
                in_features=10,
                out_features=10,
                total_bits=total_width,
                frac_bits=frac_width,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


@pytest.mark.slow
@pytest.mark.parametrize("total_bits, frac_bits, num_steps", [(4, 3, 4)])
def test_trainable_layer_linear(
    total_bits: int, frac_bits: int, num_steps: int
) -> None:
    params = FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    vrange = (params.minimum_as_rational, params.maximum_as_rational)

    stimuli = torch.rand((4, 10)) * (vrange[1] - vrange[0]) + vrange[0]
    model = SimpleLinear(total_bits, frac_bits)
    criterion = torch.nn.MSELoss()
    optim = torch.optim.Adam(model.parameters())

    for epoch in range(2):
        y = model(stimuli).squeeze()
        loss = criterion(y, torch.zeros_like(stimuli).float())
        optim.zero_grad()
        loss.backward()
        optim.step()


class SimpleBatchNormLinear(torch.nn.Module):
    def __init__(self, total_width: int, frac_width: int):
        super().__init__()
        self.model = Sequential(
            nn_fxp.BatchNormedLinear(
                in_features=10,
                out_features=10,
                total_bits=total_width,
                frac_bits=frac_width,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


@pytest.mark.slow
@pytest.mark.parametrize("total_bits, frac_bits, num_steps", [(4, 3, 4)])
def test_trainable_layer_batchnorm_linear(
    total_bits: int, frac_bits: int, num_steps: int
) -> None:
    params = FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    vrange = (params.minimum_as_rational, params.maximum_as_rational)

    stimuli = torch.rand((4, 10)) * (vrange[1] - vrange[0]) + vrange[0]

    model = SimpleBatchNormLinear(total_bits, frac_bits)
    criterion = torch.nn.MSELoss()
    optim = torch.optim.Adam(model.parameters())

    for epoch in range(2):
        y = model(stimuli).squeeze()
        loss = criterion(y, torch.zeros_like(stimuli).float())
        optim.zero_grad()
        loss.backward()
        optim.step()



from collections.abc import Callable, Iterator, Sequence
from itertools import chain

from elasticai.creator.ir2vhdl import Implementation, type_handler

from .combinatorial import wrap_in_architecture
from .language import Port, VHDLEntity
from .vhdl_nodes.node_factory import InstanceFactoryForCombinatorial
from .wiring import connect_data_signals


@type_handler
def unclocked_combinatorial(impl: Implementation) -> tuple[str, Sequence[str]]:
    def _generate_entity(name, input_size, output_size):
        entity = VHDLEntity(
            name=name,
            port=Port(
                inputs=dict(
                    d_in=f"std_logic_vector({input_size} - 1 downto 0)",
                ),
                outputs=dict(
                    d_out=f"std_logic_vector({output_size} - 1 downto 0)",
                ),
            ),
            generics=dict(),
        )
        return entity

    return impl.name, combinatorial(impl, _generate_entity)


def combinatorial(
    impl: Implementation, entity_fn: Callable[[str, int, int], VHDLEntity]
) -> Iterator[str]:
    name = impl.name
    nodes = []
    for n in impl.nodes.values():
        nodes.append(InstanceFactoryForCombinatorial(n))
    nodes = tuple(nodes)
    instances = tuple(n for n in nodes if n.name not in ("input", "output"))
    input = impl.nodes["input"]
    output = impl.nodes["output"]

    connections = tuple((e.src, e.dst, e.src_dst_indices) for e in impl.edges.values())
    signal_defs = (line for n in nodes for line in n.define_signals())

    instantiations = (line for n in instances for line in n.instantiate())
    data_signal_connections = connect_data_signals(connections)
    entity = entity_fn(name, input.input_shape.size(), output.output_shape.size())

    yield from entity.generate_entity()

    yield ""
    declarations = signal_defs
    implementation = chain(data_signal_connections, instantiations)
    yield from wrap_in_architecture(name, declarations, implementation)



from string import Template as _pyTemplate

from elasticai.creator.template import (
    AnalysingTemplateParameter,
    TemplateBuilder,
    TemplateParameter,
)


class EntityTemplateParameter(AnalysingTemplateParameter):
    """Find the entity in vhdl code and make its name available as a template paramter.

    Assumes that there is only one entity in the provided prototype.

    NOTE: Will only replace occurences where `name` is followed and preceded by a non-word character, ie.
    for an entity named `'skeleton'` the occurence `'architecture rtl of skeleton is'` will be
    replaced by `'architecture rtl of $entity is'` but the occurences `'skeleton_pkg'`,
    `'skeleton;'`, `'skeleton-'` or `skeleton.` will remain unaltered.
    IMPORTANT: This detail is likely to change in the future.
    """

    def __init__(self):
        self.name = "entity"
        self.analyse_regex = (
            r"(?i:(?<=entity )(?P<param>[a-zA-Z][a-zA-Z0-9_]*)(?=\s+is))"
        )
        self.regex = "<none>"

    def analyse(self, m: dict[str, str]) -> None:
        original_name = m["param"]
        self.regex = r"\b{original_name}\b".format(original_name=original_name)

    def replace(self, m: dict[str, str]) -> str:
        return "$entity"


class ValueTemplateParameter(TemplateParameter):
    """Find a value definition and make it settable via a template parameter.

    Searches for vhdl value definitions of the form `identifier : type`
    or `identifier : type :=` and transforms them into `identifier : type := $identifier`.

    Essentially allows to replace generics as well as variable or signal initializations.
    """

    def __init__(self, name: str):
        self.regex = (
            r"(?P<def>(?i:{name}\s*:\s*(natural|integer)))\s*(:=\s*.*)?\b".format(
                name=name
            )
        )
        self.name = name

    def replace(self, match: dict[str, str]) -> str:
        return f"{match['def']} := ${self.name}"


class EntityTemplateDirector:
    def __init__(self):
        self._builder = TemplateBuilder()
        self._builder.add_parameter(EntityTemplateParameter())

    def set_prototype(self, prototype: str) -> "EntityTemplateDirector":
        self._builder.set_prototype(prototype)
        return self

    def add_generic(self, name: str) -> "EntityTemplateDirector":
        if name == "entity":
            raise ValueError("name 'entity' is reserved for entity parameter")
        self._builder.add_parameter(ValueTemplateParameter(name))
        return self

    def build(self) -> _pyTemplate:
        return _pyTemplate(self._builder.build())



from abc import abstractmethod
from collections.abc import Callable, Collection, Hashable, Iterable, Iterator, Mapping
from typing import Any, Protocol, Self

from elasticai.creator.graph import (
    Graph,
    find_all_subgraphs,
    get_rewriteable_matches,
    rewrite,
)


class Node(Protocol):
    @property
    @abstractmethod
    def data(self) -> dict[str, Any]: ...

    @data.setter
    @abstractmethod
    def data(self, data: dict[str, Any]) -> None: ...

    @property
    @abstractmethod
    def name(self) -> str: ...


class Edge(Protocol):
    @property
    @abstractmethod
    def data(self) -> dict[str, Any]: ...

    @data.setter
    @abstractmethod
    def data(self, data: dict[str, Any]) -> None: ...
    @property
    @abstractmethod
    def src(self) -> str: ...

    @property
    @abstractmethod
    def dst(self) -> str: ...


class ReadOnlyDataGraph[N: Node, E: Edge](Protocol):
    @property
    def graph(self) -> Graph[str]: ...

    @property
    @abstractmethod
    def nodes(self) -> Mapping[str, N]: ...

    @property
    @abstractmethod
    def edges(self) -> Mapping[tuple[str, str], E]: ...

    @abstractmethod
    def successors(self, node: str) -> Mapping[str, N]: ...

    @abstractmethod
    def predecessors(self, node: str) -> Mapping[str, N]: ...

    @property
    @abstractmethod
    def data(self) -> dict[str, Any]: ...


class DataGraph[N: Node, E: Edge](ReadOnlyDataGraph[N, E], Protocol):
    @abstractmethod
    def add_node(self, node: N) -> Self: ...

    @abstractmethod
    def add_edge(self, edge: E) -> Self: ...

    @property
    @abstractmethod
    def data(self) -> dict[str, Any]: ...

    @data.setter
    @abstractmethod
    def data(self, data: dict[str, Any]) -> None: ...


class RemappedSubImplementation[N: Node, E: Edge](ReadOnlyDataGraph[N, E]):
    """Use a given mapping to make the data dictionary accessible via node names from graph.

    This allows to access the data dictionary using the names from the graph.
    The provided mapping is used to map the node names of the graph to the node names in the data dictionary.
    This is used e.g., to initialize new nodes, that were replaced during rewriting.
    The mapping is assumed to be one-to-one, i.e. each node name in the graph maps to exactly one node name in the data dictionary.

    """

    def __init__(
        self,
        mapping: dict[str, str],
        graph: Graph[str],
        data: dict[str, Any],
        node_fn: Callable[[str, dict[str, Any]], N],
        edge_fn: Callable[[str, str, dict[str, Any]], E],
    ):
        self._mapping = mapping
        self._inverted = dict(((v, k) for k, v in mapping.items()))
        self._data = data
        self._graph = graph
        self._node_constr = node_fn
        self._edge_constr = lambda src_dst, d: edge_fn(src_dst[0], src_dst[1], d)

    @property
    def graph(self) -> Graph[str]:
        return self._graph

    @property
    def data(self) -> dict[str, Any]:
        return self._data

    def _original_name(self, node: str):
        return self._mapping[node]

    def _mapped_name(self, node: str):
        return self._inverted[node]

    @property
    def nodes(self) -> Mapping[str, N]:
        return self._create_remapped_nodes(self._graph.nodes)

    @property
    def edges(self) -> Mapping[tuple[str, str], E]:
        remapping = {
            (self._original_name(src), self._original_name(dst)): (src, dst)
            for src, dst in self._graph.iter_edges()
        }
        return _RemappedData(
            remapping=remapping,
            data=self._data["edges"],
            iterable=self._graph.iter_edges(),
            factory_fn=self._edge_constr,
        )

    def successors(self, node: str) -> Mapping[str, N]:
        return self._create_remapped_nodes(
            self._graph.successors[self._inverted[node]],
        )

    def _create_remapped_nodes(self, iterable: Iterable[str]) -> Mapping[str, N]:
        return _RemappedData(
            self._mapping, self._data, iterable, factory_fn=self._node_constr
        )

    def predecessors(self, node: str) -> Mapping[str, N]:
        return self._create_remapped_nodes(
            self._graph.predecessors[self._inverted[node]]
        )


class _RemappedData[K: Hashable, D](Mapping[K, D]):
    def __init__(
        self,
        remapping: dict[K, K],
        data: dict[K, Any],
        iterable: Iterable[K],
        factory_fn: Callable[[K, dict[str, Any]], D],
    ) -> None:
        self._remapping = remapping
        self._data = data
        self._iter = iterable
        self._create_item = factory_fn

    def __getitem__(self, key) -> D:
        return self._create_item(key, self._data[self._remapping[key]])

    def __len__(self) -> int:
        return len(self._remapping)

    def __iter__(self) -> Iterator[Any]:
        for key in iter(self._iter):
            if key in self:
                yield key

    def __contains__(self, key: Any) -> bool:
        return key in self._remapping


type Rule = Callable[[DataGraph], DataGraph]


class CompositeRule:
    def __init__(self, rules: Iterable[Rule]) -> None:
        self._rules = list(rules)

    def __call__(self, impl: DataGraph) -> DataGraph:
        for rule in self._rules:
            impl = rule(impl)
        return impl


class Pattern(Protocol):
    @property
    @abstractmethod
    def graph(self) -> DataGraph: ...

    @property
    def interface(self) -> Collection[str]: ...

    @abstractmethod
    def match(self, g: ReadOnlyDataGraph) -> list[dict[str, str]]: ...


class StdPattern(Pattern):
    def __init__(
        self,
        graph: DataGraph,
        node_constraint: Callable[[Node, Node], bool],
        interface: Collection[str],
    ):
        self._pattern = graph
        self._node_constraint = node_constraint
        self._interface = interface

    @property
    def graph(self) -> DataGraph:
        return self._pattern

    @property
    def interface(self) -> Collection[str]:
        return self._interface

    def match(self, g: ReadOnlyDataGraph) -> list[dict[str, str]]:
        matches = find_all_subgraphs(
            graph=g.graph,
            pattern=self.graph.graph,
            node_constraint=lambda p_node, g_node: self._node_constraint(
                self._pattern.nodes[p_node],
                g.nodes[g_node],
            ),
        )
        matches = list(
            get_rewriteable_matches(
                matches=matches, original=g.graph, interface_nodes=set(self._interface)
            )
        )
        return matches


class IrFactory[N: Node, E: Edge, G: DataGraph](Protocol):
    @abstractmethod
    def node(self, name: str, data: Mapping[str, Any]) -> N: ...

    @abstractmethod
    def edge(self, src: str, dst: str, data: Mapping[str, Any]) -> E: ...

    @abstractmethod
    def data_graph(self) -> G: ...


class PatternRule[N: Node, E: Edge, G: DataGraph]:
    def __init__(self, spec: "PatternRuleSpec", ir_factory: IrFactory[N, E, G]):
        self._spec = spec
        self._ir_factory = ir_factory

    def __call__(self, graph: DataGraph) -> DataGraph:
        return self.apply(graph)

    def apply(self, graph: ReadOnlyDataGraph) -> DataGraph:
        self._validate_pattern(self._spec.pattern)
        matches = list(self._spec.pattern.match(graph))

        matched_subgraphs = [
            RemappedSubImplementation[N, E](
                mapping=match,
                graph=self._spec.pattern.graph.graph,
                data=graph.data["nodes"],
                node_fn=self._ir_factory.node,
                edge_fn=self._ir_factory.edge,
            )
            for match in matches
        ]
        replacements = [
            self._spec.create_replacement(matched) for matched in matched_subgraphs
        ]
        self._validate_replacements(replacements)
        new_graph, replacement_maps = self._rewrite_raw_graph_for_matches(
            graph, matches, replacements
        )
        return self._build_data_graph(
            new_raw_graph=new_graph,
            original=graph,
            replacement_map=replacement_maps,
        )

    def _build_data_graph(
        self,
        new_raw_graph: Graph[str],
        original: ReadOnlyDataGraph,
        replacement_map: dict[str, tuple[str, ReadOnlyDataGraph]],
    ) -> DataGraph:
        impl = self._ir_factory.data_graph()

        combined_map = replacement_map | {
            name: (name, original)
            for name in original.nodes
            if name in new_raw_graph.nodes and name
        }

        def copy_node(name):
            old_name, graph = combined_map[name]
            impl.add_node(
                self._ir_factory.node(name=name, data=graph.nodes[old_name].data)
            )

        def edge_lies_within_replacement(src_dst):
            src, dst = src_dst
            if src in replacement_map and dst in replacement_map:
                return True
            return False

        def copy_edge(src_dst):
            src, dst = src_dst
            if edge_lies_within_replacement(src_dst):
                old_src_name, graph = replacement_map[src]
                old_dst_name, _ = replacement_map[dst]
            else:
                old_src_name, graph = src, original
                old_dst_name, _ = dst, original

            impl.add_edge(
                self._ir_factory.edge(
                    src=src,
                    dst=dst,
                    data=graph.edges[(old_src_name, old_dst_name)].data,
                )
            )

        for node_name in new_raw_graph.nodes:
            copy_node(node_name)

        for src, dst in new_raw_graph.iter_edges():
            copy_edge((src, dst))

        return impl

    def _validate_replacements(self, replacements: list[ReadOnlyDataGraph]) -> None:
        for replacement in replacements:
            missing_interface_nodes = set(self._spec.interface) - set(
                replacement.nodes.keys()
            )
            if missing_interface_nodes:
                raise ValueError(
                    f"Replacement is missing interface nodes: {missing_interface_nodes}"
                )

    def _validate_pattern(self, pattern: Pattern) -> None:
        missing_interface_nodes = set(self._spec.interface) - set(
            pattern.graph.nodes.keys()
        )
        if missing_interface_nodes:
            raise ValueError(
                f"Pattern Graph is missing interface nodes: {missing_interface_nodes}"
            )

    def _rewrite_raw_graph_for_matches(
        self,
        graph: ReadOnlyDataGraph,
        matches: list[dict[str, str]],
        replacements: list[ReadOnlyDataGraph],
    ) -> tuple[Graph[str], dict[str, tuple[str, ReadOnlyDataGraph]]]:
        rewritten = graph.graph
        full_replacement_map = {}
        for match, replacement in zip(matches, replacements):
            rewritten, replacement_map = rewrite(
                replacement=replacement.graph,
                original=rewritten,
                match=match,
                lhs={x: x for x in self._spec.interface},
                rhs={x: x for x in self._spec.interface},
            )
            for replacement_node_name, new_node_name in replacement_map.items():
                full_replacement_map[new_node_name] = (
                    replacement_node_name,
                    replacement,
                )

        return rewritten, full_replacement_map


class PatternRuleSpec:
    """Specifies how `PatternRule` should create a new `DataGraph` from an existing one.

    It consists of

    - a pattern to match in the original DataGraph.
    - a replacement to apply when the pattern is matched. The
    `PatternRule` will use the `replacement_fn` function to create a
    new `DataGraph`. The function receives the matched
    part of the DataGraph as an argument, so you can build
    the replacement based on the parameters found in your
    matched pattern. The node names in this matched
    subimplementation are remapped from the original graph to
    the pattern graph, so you can access the data dictionary
    using the names from the pattern graph. E.g., if the
    pattern specifies a node `'conv'` you can acces the data of
    the original DataGraph that this node corresponds to
    using `matched.nodes['conv'].data`.
    - an interface that specifies which nodes are part of the
    pattern and of the replacement, nodes that are part of the
    interface are neither replaced nor initialized.
    Instead they are used to connect the pattern and the
    replacement.
    """

    def __init__(
        self,
        pattern: Pattern,
        replacement_fn: Callable[[ReadOnlyDataGraph], ReadOnlyDataGraph],
    ):
        self.pattern = pattern
        self.create_replacement = replacement_fn

    @property
    def interface(self) -> Collection[str]:
        return self.pattern.interface



import torch

from elasticai.creator.nn.fixed_point.conv1d.layer.batch_normed_conv1d import (
    BatchNormedConv1d,
)


def conv1d(signal_length: int, bias: bool, affine: bool) -> BatchNormedConv1d:
    return BatchNormedConv1d(
        total_bits=4,
        frac_bits=1,
        kernel_size=2,
        signal_length=signal_length,
        bias=bias,
        in_channels=1,
        out_channels=2,
        bn_affine=affine,
        bn_momentum=1,
    )


def test_output_contains_correct_number_of_output_channels() -> None:
    conv = conv1d(signal_length=15, bias=False, affine=False)
    input_data = torch.rand(1, 15)
    prediction = conv(input_data)
    num_channels, _ = prediction.shape
    print(prediction)
    assert num_channels == 2


def test_that_batch_dimension_is_kept() -> None:
    conv = conv1d(signal_length=15, bias=False, affine=False)
    input_data = torch.rand(3, 1, 15)
    prediction = conv(input_data)
    print(prediction.shape)
    batch_dimension = prediction.shape[0]
    assert batch_dimension == 3



import pytest
import torch
from torch.nn import Sequential
from torch.optim import SGD

from elasticai.creator_plugins.quantized_grads.base_modules import Linear
from elasticai.creator_plugins.quantized_grads.fixed_point import (
    FixedPointConfigV2,
    QuantizeForwHTEBackwHTE,
    QuantizeParamToFixedPointHTE,
)
from elasticai.creator_plugins.quantized_grads.quantized_optim import (
    get_quantized_optimizer,
)


@pytest.mark.slow
class Test1:
    def test_pred(self, pred: torch.Tensor) -> None:
        expected_pred = torch.Tensor([0.5, 2])
        assert torch.equal(pred, expected_pred)

    def test_loss(self, loss):
        expected_loss = 0.6250
        assert loss == expected_loss

    def test_grad_weight(self, model, loss):
        expected_grad_weight = torch.Tensor([[0, 0, 0], [0, 1, 2]])
        actual_grad_weight = model[0].weight.grad
        assert torch.equal(actual_grad_weight, expected_grad_weight)

    def test_grad_bias(self, model, loss):
        expected_grad_bias = torch.Tensor([0, 1])
        actual_grad_bias = model[0].bias.grad
        assert torch.equal(actual_grad_bias, expected_grad_bias)

    def test_new_weight(self, model, optimizer_step):
        expected_new_weight = torch.Tensor([[0, 0.125, 0.25], [0.375, 0.375, 0.375]])
        actual_new_weight = model[0].weight
        assert torch.equal(actual_new_weight, expected_new_weight)

    def test_new_bias(self, model, optimizer_step):
        expected_new_bias = torch.Tensor([0, 0])
        actual_new_bias = model[0].bias
        assert torch.equal(actual_new_bias, expected_new_bias)

    @pytest.fixture(scope="class")
    def in_features(self) -> int:
        in_features = 3
        return in_features

    @pytest.fixture(scope="class")
    def out_features(self) -> int:
        out_features = 2
        return out_features

    @pytest.fixture(scope="class")  # ensures expensive setup is performed only once
    def model(self, in_features: int, out_features: int) -> Sequential:
        params_conf_total_bits = 8
        params_conf_frac_bits = 3
        params_conf = FixedPointConfigV2(
            total_bits=params_conf_total_bits, frac_bits=params_conf_frac_bits
        )

        forward_conf_total_bits = 8
        forward_conf_frac_bits = 2
        forward_conf = FixedPointConfigV2(
            total_bits=forward_conf_total_bits,
            frac_bits=forward_conf_frac_bits,
        )

        grad_conf_total_bits = 8
        grad_conf_frac_bits = 0
        backward_conf = FixedPointConfigV2(
            total_bits=grad_conf_total_bits,
            frac_bits=grad_conf_frac_bits,
        )

        nn = Sequential(
            Linear(
                math_ops=QuantizeForwHTEBackwHTE(
                    forward_conf=forward_conf, backward_conf=backward_conf
                ),
                in_features=in_features,
                out_features=out_features,
                weight_quantization=QuantizeParamToFixedPointHTE(params_conf),
                bias_quantization=QuantizeParamToFixedPointHTE(params_conf),
                bias=True,
            )
        )
        nn[0].weight = torch.reshape(
            input=torch.arange(
                start=0,
                end=(out_features * in_features) / (2**params_conf_frac_bits),
                step=1 / (2**params_conf_frac_bits),
                requires_grad=True,
                dtype=torch.float32,
            ),
            shape=nn[0].weight.shape,
        )
        nn[0].bias = torch.reshape(
            input=torch.arange(
                start=0,
                end=out_features / (2**params_conf_frac_bits),
                step=1 / (2**params_conf_frac_bits),
                dtype=torch.float32,
            ),
            shape=nn[0].bias.shape,
        )
        return nn

    @pytest.fixture(scope="class")
    def input(self, in_features: int) -> torch.Tensor:
        return torch.arange(start=0, end=in_features, dtype=torch.float32)

    @pytest.fixture(scope="class")
    def pred(self, model: Sequential, input: torch.Tensor) -> torch.Tensor:
        return model(input)

    @pytest.fixture(scope="class")
    def output(self, out_features: int) -> torch.Tensor:
        return torch.arange(start=0, end=out_features, dtype=torch.float32)

    @pytest.fixture(scope="class")
    def loss(self, output: torch.Tensor, pred: torch.Tensor) -> torch.Tensor:
        loss_fn = torch.nn.MSELoss()
        _loss = loss_fn(pred, output)
        _loss.backward()
        return _loss

    @pytest.fixture(scope="class")
    def optimizer(self, model: Sequential, loss: torch.Tensor) -> torch.optim.Optimizer:
        QSGD = get_quantized_optimizer(SGD)
        optimizer = QSGD(
            model=model,
            params=model.parameters(),
            lr=0.1,
            momentum=0.9,
        )
        return optimizer

    @pytest.fixture(scope="class")
    def optimizer_step(self, optimizer: SGD, loss: torch.Tensor) -> None:
        optimizer.step()


@pytest.mark.slow
class Test2:
    def test_pred(self, pred: torch.Tensor) -> None:
        expected_pred = torch.Tensor([1.25, 3.75])
        assert torch.equal(pred, expected_pred)

    def test_loss(self, loss):
        expected_loss = 4.5625
        assert loss == expected_loss

    def test_grad_weight(self, model, loss):
        expected_grad_weight = torch.Tensor([[0, 1.25, 2.5], [0, 2.75, 5.5]])
        actual_grad_weight = model[0].weight.grad
        assert torch.equal(actual_grad_weight, expected_grad_weight)

    def test_grad_bias(self, model, loss):
        expected_grad_bias = torch.Tensor([1.25, 2.75])
        actual_grad_bias = model[0].bias.grad
        assert torch.equal(actual_grad_bias, expected_grad_bias)

    def test_new_weight(self, model, optimizer_step):
        expected_new_weight = torch.Tensor([[0, 0, 0.25], [0.75, 0.75, 0.75]])
        actual_new_weight = model[0].weight
        assert torch.equal(actual_new_weight, expected_new_weight)

    def test_new_bias(self, model, optimizer_step):
        expected_new_bias = torch.Tensor([0, 0])
        actual_new_bias = model[0].bias
        assert torch.equal(actual_new_bias, expected_new_bias)

    @pytest.fixture(scope="class")
    def in_features(self) -> int:
        in_features = 3
        return in_features

    @pytest.fixture(scope="class")
    def out_features(self) -> int:
        out_features = 2
        return out_features

    @pytest.fixture(scope="class")  # ensures expensive setup is performed only once
    def model(self, in_features: int, out_features: int) -> Sequential:
        params_conf_total_bits = 8
        params_conf_frac_bits = 2
        params_conf = FixedPointConfigV2(
            total_bits=params_conf_total_bits, frac_bits=params_conf_frac_bits
        )

        forward_conf_total_bits = 8
        forward_conf_frac_bits = 4
        forward_conf = FixedPointConfigV2(
            total_bits=forward_conf_total_bits,
            frac_bits=forward_conf_frac_bits,
        )

        grad_conf_total_bits = 8
        grad_conf_frac_bits = 4
        backward_conf = FixedPointConfigV2(
            total_bits=grad_conf_total_bits,
            frac_bits=grad_conf_frac_bits,
        )

        nn = Sequential(
            Linear(
                math_ops=QuantizeForwHTEBackwHTE(
                    forward_conf=forward_conf, backward_conf=backward_conf
                ),
                in_features=in_features,
                out_features=out_features,
                weight_quantization=QuantizeParamToFixedPointHTE(params_conf),
                bias_quantization=QuantizeParamToFixedPointHTE(params_conf),
                bias=True,
            )
        )
        nn[0].weight = torch.reshape(
            input=torch.arange(
                start=0,
                end=(out_features * in_features) / (2**params_conf_frac_bits),
                step=1 / (2**params_conf_frac_bits),
                requires_grad=True,
                dtype=torch.float32,
            ),
            shape=nn[0].weight.shape,
        )
        nn[0].bias = torch.reshape(
            input=torch.arange(
                start=0,
                end=out_features / (2**params_conf_frac_bits),
                step=1 / (2**params_conf_frac_bits),
                dtype=torch.float32,
            ),
            shape=nn[0].bias.shape,
        )
        return nn

    @pytest.fixture(scope="class")
    def input(self, in_features: int) -> torch.Tensor:
        return torch.arange(start=0, end=in_features, dtype=torch.float32)

    @pytest.fixture(scope="class")
    def pred(self, model: Sequential, input: torch.Tensor) -> torch.Tensor:
        return model(input)

    @pytest.fixture(scope="class")
    def output(self, out_features: int) -> torch.Tensor:
        return torch.arange(start=0, end=out_features, dtype=torch.float32)

    @pytest.fixture(scope="class")
    def loss(self, output: torch.Tensor, pred: torch.Tensor) -> torch.Tensor:
        loss_fn = torch.nn.MSELoss()
        _loss = loss_fn(pred, output)
        _loss.backward()
        return _loss

    @pytest.fixture(scope="class")
    def optimizer(self, model: Sequential, loss: torch.Tensor) -> torch.optim.Optimizer:
        QSGD = get_quantized_optimizer(SGD)
        optimizer = QSGD(
            model=model,
            params=model.parameters(),
            lr=0.1,
            momentum=0.9,
        )
        return optimizer

    @pytest.fixture(scope="class")
    def optimizer_step(self, optimizer: SGD, loss: torch.Tensor) -> None:
        optimizer.step()



import pytest
import torch

from elasticai.creator.arithmetic import FxpParams
from elasticai.creator.nn import fixed_point as nn_creator


@pytest.mark.parametrize(
    "total_bits, frac_bits, num_steps",
    [(4, 3, 4), (6, 4, 32), (8, 5, 64), (10, 6, 16), (12, 7, 128), (12, 11, 64)],
)
def test_relu_compared_torch(total_bits: int, frac_bits: int, num_steps: int) -> None:
    fxp = FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    vrange = (fxp.minimum_as_rational, fxp.maximum_as_rational)
    stimulus = torch.arange(
        start=vrange[0], end=vrange[1], step=fxp.minimum_step_as_rational
    )

    act0 = torch.nn.ReLU()
    out0 = act0(stimulus)
    act1 = nn_creator.ReLU(total_bits)
    out1 = act1(stimulus)
    act2 = nn_creator.ReLU(total_bits, True)
    out2 = act2(stimulus)

    assert float(sum(abs(out1 - out0))) < 1e-6
    assert float(sum(abs(out2 - out0))) < 1e-6



from collections.abc import Iterator
from typing import TypeVar

from elasticai.creator.graph.graph import Graph

from .._types import NodeConstraintFn
from .state import State

T = TypeVar("T")
TP = TypeVar("TP")


class MatchError(Exception):
    def __init__(self):
        super().__init__("No match found")


def match(
    pattern: Graph[TP],
    graph: Graph[T],
    node_constraint: NodeConstraintFn[TP, T] = lambda _, __: True,
) -> dict[TP, T]:
    try:
        return next(_match(pattern, graph, node_constraint))
    except StopIteration:
        raise MatchError()


def find_all_matches(
    pattern: Graph[TP],
    graph: Graph[T],
    node_constraint: NodeConstraintFn[TP, T] = lambda _, __: True,
) -> list[dict[TP, T]]:
    return list(_match(pattern, graph, node_constraint))


def _match(
    pattern: Graph[TP],
    graph: Graph[T],
    node_constraint: NodeConstraintFn[TP, T] = lambda _, __: True,
) -> Iterator[dict[TP, T]]:
    """
    Implementation of the VF2 algorithm by Cordella et al. 2004
    """
    pattern_state: State[TP, T] = State(
        graph=pattern,
    )
    graph_state: State[T, TP] = State(
        graph=graph,
    )

    def compute_candidate_pairs(
        pattern_state: State[TP, T],
        graph_state: State[T, TP],
        depth: int,
    ) -> Iterator[tuple[TP, T]]:
        candidate_pairs: set[tuple[TP, T]] = set()
        out_of_g: set[T] = set(graph_state.iter_out_nodes())
        out_of_p: set[TP] = set(pattern_state.iter_out_nodes())
        if len(out_of_g) == 0 or len(out_of_p) == 0:
            into_g: set[T] = set(graph_state.iter_in_nodes())
            into_p: set[TP] = set(pattern_state.iter_in_nodes())
            for pn in into_p:
                for gn in into_g:
                    candidate_pairs.add((pn, gn))

        else:
            for pn in out_of_p:
                for gn in out_of_g:
                    candidate_pairs.add((pn, gn))

        if len(candidate_pairs) == 0 and depth == 0:
            for pn in pattern.nodes:
                for gn in graph.nodes:
                    yield pn, gn
        else:
            yield from candidate_pairs

    def is_feasible(pn: TP, gn: T) -> bool:
        partial_pattern_pred = pattern_state.partial_predecessors(pn)
        partial_pattern_succ = pattern_state.partial_successors(pn)
        partial_graph_pred = graph_state.partial_predecessors(gn)
        partial_graph_succ = graph_state.partial_successors(gn)

        pattern_in_node_pred = pattern_state.in_node_predecessors(pn)
        pattern_out_node_pred = pattern_state.out_node_predecessors(pn)
        pattern_in_node_succ = pattern_state.in_node_successors(pn)
        pattern_out_node_succ = pattern_state.out_node_successors(pn)
        pattern_unseen_pred = pattern_state.unseen_predecessors(pn)
        pattern_unseen_succ = pattern_state.unseen_successors(pn)

        full_graph_pred = graph.predecessors[gn]
        full_graph_succ = graph.successors[gn]
        full_pattern_pred = pattern.predecessors[pn]
        full_pattern_succ = pattern.successors[pn]

        graph_in_node_pred = graph_state.in_node_predecessors(gn)
        graph_out_node_pred = graph_state.out_node_predecessors(gn)

        graph_in_node_succ = graph_state.in_node_successors(gn)
        graph_out_node_succ = graph_state.out_node_successors(gn)

        graph_unseen_pred = graph_state.unseen_predecessors(gn)
        graph_unseen_succ = graph_state.unseen_successors(gn)

        def was_matched(ps, gs):
            return pattern_state.contains_pair(ps, gs)

        def predecessors_check():
            """
            ensure that each predecessors of the considered nodes
            that is in our partial pattern match (pattern_state.core) also has a corresponding
            node in our partial graph match (graph_state.core)
            and vice versa.
            This also prevents a node from being matched if including it would include new edges in the match, that are not present in the pattern, e.g.,
            consider the graph

            a: {b, c}
            b: {c}

            would not match the pattern
            a: {b}
            b: {c}

            because including c would introduce the new edge (a, c) which is not present in the pattern.
            """
            for ps in partial_pattern_pred:
                found = False
                for gs in full_graph_pred:
                    if was_matched(ps, gs):
                        found = True
                        break
                if not found:
                    return False

            for gs in partial_graph_pred:
                found = False
                for ps in full_pattern_pred:
                    if was_matched(ps, gs):
                        found = True
                        break
                if not found:
                    return False
            return True

        def successors_check():
            """
            analogue to predecessor check
            """
            for ps in partial_pattern_succ:
                found = False
                for gs in full_graph_succ:
                    if was_matched(ps, gs):
                        found = True
                        break
                if not found:
                    return False
            for gs in partial_graph_succ:
                found = False
                for ps in full_pattern_succ:
                    if was_matched(ps, gs):
                        found = True
                        break
                if not found:
                    return False
            return True

        def in_check():
            "checks nodes that are might be added"
            return len(graph_in_node_pred) >= len(pattern_in_node_pred) and len(
                graph_in_node_succ
            ) >= len(pattern_in_node_succ)

        def out_check():
            return len(graph_out_node_pred) >= len(pattern_out_node_pred) and len(
                graph_out_node_succ
            ) >= len(pattern_out_node_succ)

        def new_check():
            """Considers nodes that might be added in current_depth + 2"""
            return len(graph_unseen_pred) >= len(pattern_unseen_pred) and len(
                graph_unseen_succ
            ) >= len(pattern_unseen_succ)

        _in_check = in_check()

        _out_check = out_check()

        _predecessor_check = predecessors_check()

        _successors_check = successors_check()

        _new_check = new_check()

        _constraint = node_constraint(pn, gn)

        result = (
            _in_check
            and _out_check
            and _predecessor_check
            and _successors_check
            and _new_check
            and _constraint
        )

        return result

    def do_match(depth: int) -> Iterator[dict[TP, T]]:
        if pattern_state.is_complete():
            yield dict(pattern_state.iter_matched_pairs())
            return
        pattern_state.next_depth()
        graph_state.next_depth()
        candidate_pairs = list(
            compute_candidate_pairs(pattern_state, graph_state, depth=depth)
        )

        order = {pn: i for i, (pn, _) in enumerate(candidate_pairs)}
        for pn, gn in candidate_pairs:
            if any(n in pattern.nodes and order[n] < order[pn] for n in order):
                continue
            if is_feasible(pn, gn):
                pattern_state.add_pair(pn, gn)
                graph_state.add_pair(gn, pn)
                yield from do_match(depth=depth + 1)
                pattern_state.remove_pair(pn, gn)
                graph_state.remove_pair(gn, pn)

        pattern_state.restore()
        graph_state.restore()

    yield from do_match(depth=0)



from unittest import TestCase

from elasticai.creator_plugins.grouped_filter import FilterParameters


class GetInChannelsByGroup(TestCase):
    def test_two_in_channels_in_two_groups_gets_0_1(self):
        p = FilterParameters(kernel_size=1, in_channels=2, out_channels=2, groups=2)
        self.assertEqual(((0,), (1,)), p.get_in_channels_by_group())

    def test_4_in_channels_2_groups_gets_01_23(self):
        p = FilterParameters(kernel_size=1, in_channels=4, out_channels=4, groups=2)
        self.assertEqual(((0, 1), (2, 3)), p.get_in_channels_by_group())



from elasticai.creator.ir2vhdl import Shape, edge, vhdl_node
from elasticai.creator_plugins.grouped_filter import FilterParameters
from elasticai.creator_plugins.time_multiplexed_sequential.src import (
    sequential,
)

from .high_level_network import network


def test_inserts_edge_from_input_to_sliding_window():
    nw = network(
        input_shape=(1, 4),
        kernel_size=3,
        out_channels=1,
        stride=2,
    )
    lowered = sequential(nw)
    expected = [
        edge(
            src="input",
            dst="conv0_i0",
            src_dst_indices=tuple(),
        )
    ]
    actual = []
    for e in lowered.edges.values():
        if e.src == "input":
            actual.append(e)

    assert actual == expected


def test_inserts_striding_shift_register():
    kernel_size = 3
    out_channels = 1
    nw = network(
        kernel_size=kernel_size,
        out_channels=out_channels,
        input_shape=(1, 10),
        stride=2,
    )
    lowered = sequential(nw)
    type = "striding_shift_register"
    name = f"{type}_i0"
    expected = vhdl_node(
        name=name,
        input_shape=Shape(1, out_channels),
        output_shape=Shape(1, kernel_size),
        implementation=type,
        type=type,
        attributes={"generic_map": {"stride": 2}},
    ).data
    for n in lowered.nodes.values():
        print(n)
    assert lowered.nodes[name].data == expected


def test_convolution_is_translated_to_unclocked_combinatorial():
    kernel_size = 3
    out_channels = 1
    nw = network(
        kernel_size=kernel_size,
        out_channels=out_channels,
        input_shape=(1, 10),
        stride=2,
    )
    lowered = sequential(nw)
    assert lowered.nodes["conv0_i0"].type == "unclocked_combinatorial"


def test_inserts_conv_node():
    lowered = sequential(
        network(input_shape=(1, 10), kernel_size=3, out_channels=1, stride=1)
    )

    expected = vhdl_node(
        name="sliding_window",
        input_shape=(
            1,
            3,
        ),
        output_shape=(
            1,
            1,
        ),
        type="unclocked_combinatorial",
        implementation="conv0",
        attributes={
            "filter_parameters": FilterParameters(
                kernel_size=3,
                in_channels=1,
                out_channels=1,
                output_size=1,
                stride=1,
            ).as_dict()
        },
    )
    actual = tuple(lowered.successors("input").values())[0]
    assert actual.data == expected.data



from collections.abc import Callable
from typing import ParamSpec

import elasticai.creator.plugin as _pl
from elasticai.creator.function_utils import FunctionDecorator
from elasticai.creator.ir2vhdl import (
    Edge,
    Implementation,
    LoweringPass,
    Shape,
    edge,
    vhdl_node,
)
from elasticai.creator.ir2vhdl import (
    VhdlNode as Node,
)

from .filter_params import FilterParameters
from .index_generators import GroupedFilterIndexGenerator

P = ParamSpec("P")


def _type_handler_fn(
    name: str, fn: Callable[[Implementation], Implementation]
) -> _pl.PluginSymbol:
    def load_into(lower: LoweringPass[Implementation, Implementation]):
        lower.register(name, fn)

    return _pl.make_plugin_symbol(load_into, fn)


_type_handler = FunctionDecorator(_type_handler_fn)


def append_counter_suffix_before_construction(
    fn: Callable[P, Node],
) -> Callable[P, Node]:
    counters: dict[str, int] = {}

    def construct(*args: P.args, **kwargs: P.kwargs) -> Node:
        nonlocal counters
        if "name" in kwargs:
            name = kwargs["name"]
            if not isinstance(name, str):
                raise TypeError("expected `name` to be of type str")
        elif isinstance(args[0], str):
            name = args[0]
        else:
            raise TypeError("missing positional arg `name`")

        count = counters.get(name, 0)
        counters[name] = count + 1
        kwargs["name"] = f"{name}_i{count}"
        node = fn(*args, **kwargs)
        return node

    return construct


@_type_handler
def grouped_filter(impl: Implementation) -> Implementation:
    nc: Callable = append_counter_suffix_before_construction(vhdl_node)
    result: Implementation[Node, Edge] = Implementation(
        name=impl.name, type="unclocked_combinatorial", data={}
    )
    params = FilterParameters.from_dict(impl.attributes["filter_parameters"])
    result.add_node(
        vhdl_node(
            name="input",
            type="input",
            implementation="",
            input_shape=Shape(params.in_channels, params.kernel_size),
            output_shape=Shape(params.in_channels, params.kernel_size),
        )
    )
    kernels = impl.attributes["kernel_per_group"]
    if len(kernels) != params.groups:
        raise ValueError(
            "number of kernels per group should match the number of groups but found {}  and {}".format(
                len(kernels), params.groups
            )
        )
    g = GroupedFilterIndexGenerator(
        params=FilterParameters(
            kernel_size=params.kernel_size,
            in_channels=params.in_channels,
            out_channels=params.out_channels,
            stride=1,
            input_size=params.kernel_size,
            groups=params.groups,
        )
    )
    output_offset = 0
    for kernel, wires_per_step in zip(kernels, g.as_tuple_by_groups()):
        wires = wires_per_step[0]
        node = nc(
            name=kernel,
            type="unclocked_combinatorial",
            implementation=kernel,
            input_shape=Shape(
                len(wires),
            ),
            output_shape=Shape(
                params.out_channels_per_group,
            ),
        )
        result.add_node(node)
        result.add_edge(
            edge(
                src="input",
                dst=node.name,
                src_dst_indices=tuple(zip(wires, range(len(wires)))),
            )
        )
        result.add_edge(
            edge(
                src=node.name,
                dst="output",
                src_dst_indices=(
                    f"range(0, {params.out_channels_per_group})",
                    f"range({output_offset}, {output_offset + params.out_channels_per_group})",
                ),
            ),
        )
        output_offset += params.out_channels_per_group

    result.add_node(
        vhdl_node(
            name="output",
            type="output",
            implementation="",
            input_shape=Shape(params.out_channels, 1),
            output_shape=Shape(params.out_channels, 1),
        )
    )
    return result



from torch import Tensor, flatten

from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)


def parse_fxp_tensor_to_bytearray(
    tensor: Tensor, total_bits: int, frac_bits: int
) -> list[bytearray]:
    tensor = flatten(tensor.permute([0, 2, 1]), start_dim=1)
    fxp_params = FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    fxp_config = FxpArithmetic(fxp_params)
    ints = fxp_config.cut_as_integer(tensor).tolist()
    data = list()
    for i, batch in enumerate(ints):
        data.append(bytearray())
        for item in batch:
            item_as_bytes = int(item).to_bytes(1, byteorder="big", signed=True)
            data[i].extend(item_as_bytes)
    return data


def parse_bytearray_to_fxp_tensor(
    data: list[bytearray], total_bits: int, frac_bits: int, dimensions: tuple
) -> Tensor:
    fxp_params = FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    fxp_config = FxpArithmetic(fxp_params)
    rationals = list()
    for i, batch in enumerate(data):
        rationals.append(list())
        my_batch = batch.hex(sep=" ", bytes_per_sep=1).split(" ")
        for item in my_batch:
            item_as_bytes = bytes.fromhex(item)
            item_as_int = int.from_bytes(item_as_bytes, byteorder="big", signed=True)
            item_as_rational = fxp_config.as_rational(item_as_int)
            rationals[i].append(item_as_rational)
    tensor = Tensor(rationals)
    tensor = tensor.unflatten(1, (dimensions[2], dimensions[1]))
    tensor = tensor.transpose(1, 2)
    return tensor



from collections import OrderedDict

import torch
from torch.nn import Linear, ReLU, Sequential

from elasticai.creator.torch2ir import get_default_converter


def model():
    return Sequential(
        Linear(1, 2),
        ReLU(),
    )


def convert(model):
    ir = get_default_converter().convert(model)
    return [impl.as_dict() for impl in ir]


def test_convert_linear_without_bias():
    m = Sequential(Linear(1, 2, bias=False))
    with torch.no_grad():
        m.get_submodule("0").weight.mul_(0).add_(1)
    ir = convert(m)
    assert ir == [
        {
            "name": "",
            "type": "module",
            "nodes": {
                "input_1": {
                    "type": "input",
                    "implementation": "input",
                },
                "_0": {
                    "type": "linear",
                    "implementation": "0",
                },
                "output": {
                    "type": "output",
                    "implementation": "output",
                },
            },
            "edges": {"input_1": {"_0": dict()}, "_0": {"output": dict()}},
        },
        {
            "name": "0",
            "type": "linear",
            "in_features": 1,
            "out_features": 2,
            "bias": False,
            "edges": {},
            "nodes": {},
        },
    ]


def test_convert_linear_to_ir():
    m = model()
    with torch.no_grad():
        linear = m.get_submodule("0")
        linear.bias.mul_(0)
        linear.weight.mul_(0).add_(1)

    assert convert(m) == [
        {
            "name": "",
            "type": "module",
            "nodes": {
                "input_1": {
                    "implementation": "input",
                    "type": "input",
                },
                "_0": {
                    "implementation": "0",
                    "type": "linear",
                },
                "_1": {
                    "implementation": "1",
                    "type": "relu",
                },
                "output": {
                    "implementation": "output",
                    "type": "output",
                },
            },
            "edges": {"input_1": {"_0": {}}, "_0": {"_1": {}}, "_1": {"output": {}}},
        },
        {
            "type": "linear",
            "in_features": 1,
            "out_features": 2,
            "bias": True,
            "edges": {},
            "name": "0",
            "nodes": {},
        },
        {
            "type": "relu",
            "name": "1",
            "edges": {},
            "nodes": {},
        },
    ]


def test_converting_model_with_batchnorm():
    class Model(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.bn = torch.nn.BatchNorm1d(2)
            self.relu = torch.nn.ReLU()

        def forward(self, x):
            return self.relu(self.bn(x))

    m = Model()
    assert convert(m) == [
        {
            "edges": {"x": {"bn": {}}, "bn": {"relu": {}}, "relu": {"output": {}}},
            "name": "",
            "nodes": {
                "x": {
                    "implementation": "input",
                    "type": "input",
                },
                "bn": {
                    "implementation": "bn",
                    "type": "batchnorm1d",
                },
                "relu": {
                    "implementation": "relu",
                    "type": "relu",
                },
                "output": {
                    "implementation": "output",
                    "type": "output",
                },
            },
            "type": "module",
        },
        {
            "name": "bn",
            "type": "batchnorm1d",
            "edges": {},
            "nodes": {},
            "num_features": 2,
            "affine": True,
        },
        {
            "edges": {},
            "name": "relu",
            "nodes": {},
            "type": "relu",
        },
    ]


def test_can_handle_same_object_under_different_hierarchy_paths():
    lin = Linear(1, 1)
    model = Sequential(OrderedDict(a=lin, b=lin))
    assert convert(model) == [
        {
            "name": "",
            "type": "module",
            "nodes": {
                "input_1": {
                    "type": "input",
                    "implementation": "input",
                },
                "a": {
                    "type": "linear",
                    "implementation": "a",
                },
                "a_1": {
                    "type": "linear",
                    "implementation": "a",
                },
                "output": {
                    "type": "output",
                    "implementation": "output",
                },
            },
            "edges": {
                "input_1": {"a": {}},
                "a": {"a_1": {}},
                "a_1": {"output": {}},
            },
        },
        {
            "name": "a",
            "type": "linear",
            "in_features": 1,
            "out_features": 1,
            "bias": True,
            "edges": {},
            "nodes": {},
        },
    ]


def test_can_handle_nested_hierarchies():
    model = Sequential(OrderedDict(top=Sequential(OrderedDict(nested=Linear(1, 1)))))
    assert convert(model) == [
        {
            "name": "",
            "type": "module",
            "nodes": {
                "input_1": {
                    "type": "input",
                    "implementation": "input",
                },
                "top_nested": {
                    "type": "linear",
                    "implementation": "top.nested",
                },
                "output": {
                    "type": "output",
                    "implementation": "output",
                },
            },
            "edges": {"input_1": {"top_nested": {}}, "top_nested": {"output": {}}},
        },
        {
            "name": "top.nested",
            "type": "linear",
            "in_features": 1,
            "out_features": 1,
            "bias": True,
            "edges": {},
            "nodes": {},
        },
    ]



from collections.abc import Iterable

import pytest

from elasticai.creator.ir2vhdl import EntityTemplateDirector


@pytest.fixture
def dummy_vhdl_prototype() -> Iterable[str]:
    return [
        "library ieee;",
        "use ieee.std.numeric.all;",
        "use ieee.logic_1644.all;",
        "use work.skeleton_pkg.all;",
        "",
        "entity skeleton is",
        "  generic (",
        "    DATA_WIDTH : natural;",
        "    DATA_DEPTH : natural",
        "  );",
        "  port (",
        "    d_in: std_logic_vector(DATA_WIDTH - 1 downto 0);",
        "    address_in: std_logic_vector(16 - 1 downto 0)",
        "  );",
        "end entity;",
        "",
        "begin architecture rtl of skeleton is",
        "  constant ADDRESS_WIDTH: integer := 16;begin  d_in <= address_in;",
        "end architecture;",
    ]


class TestVhdlTemplating:
    def test_replace_entity_name(self, dummy_vhdl_prototype) -> None:
        type_handler = (
            EntityTemplateDirector().set_prototype(dummy_vhdl_prototype).build()
        )
        code = type_handler.substitute(
            {
                "entity": "my_first_hw_component",
            }
        ).splitlines()
        assert (
            "entity my_first_hw_component is",
            "begin architecture rtl of my_first_hw_component is",
        ) == (code[5], code[16])

    def test_replace_generic(self, dummy_vhdl_prototype) -> None:
        type_handler = (
            EntityTemplateDirector()
            .add_generic("data_width")
            .set_prototype("""
generic (
  DATA_WIDTH : natural
);
""")
            .build()
        )
        code = type_handler.substitute(dict(data_width=5)).splitlines()
        assert "  DATA_WIDTH : natural := 5" == code[2]

    def test_replace_first_of_two_generics(self) -> None:
        type_handler = (
            EntityTemplateDirector()
            .set_prototype("""
generic (
    DATA_WIDTH : natural;
    DATA_DEPTH : natural
    );
""")
            .add_generic("data_width")
            .build()
        )
        code = type_handler.substitute(dict(data_width=5)).splitlines()
        assert "    DATA_WIDTH : natural := 5;" == code[2]

    def test_can_change_assigned_value(self) -> None:
        type_handler = (
            EntityTemplateDirector()
            .set_prototype("""
generic (
    DATA_WIDTH : natural := 8
    );
""")
            .add_generic("data_width")
            .build()
        )
        code = type_handler.substitute(dict(data_width=5)).splitlines()
        assert "    DATA_WIDTH : natural := 5" == code[2]

    def test_can_replace_two_parameters(self) -> None:
        type_handler = (
            EntityTemplateDirector()
            .set_prototype("""
generic (
    DATA_WIDTH : natural;
    DATA_DEPTH : natural
    );
""")
            .add_generic("data_width")
            .add_generic("data_depth")
            .build()
        )
        code = type_handler.substitute(dict(data_width=5, data_depth=3)).splitlines()
        assert "    DATA_WIDTH : natural := 5;" == code[2]
        assert "    DATA_DEPTH : natural := 3" == code[3]



from collections.abc import Callable
from functools import singledispatchmethod
from re import Match, Pattern
from re import compile as _regex_compile
from typing import Iterable, Protocol, runtime_checkable


@runtime_checkable
class TemplateParameter(Protocol):
    regex: str
    """A regular expression to match what should be turned into a template parameter."""

    def replace(self, match: dict[str, str]) -> str:
        """Define how to replace the matched pattern by the template parameter.

        The match is a dictionary of named capture groups.
        In the most simple cases you can just return `f"${self.name}"` to replace the
        full match by the template parameter.
        But you can also use the match to create a more complex replacement.

        The full match is available as `match[""]`.
        """
        ...


@runtime_checkable
class AnalysingTemplateParameter(TemplateParameter, Protocol):
    """A template parameter type that needs an analysis phase before the replacement.

    You could e.g., use this to find the first class definition in a python file and
    use that class name to build the `regex` parameter and the replacement to
    replace all occurences of that class name by a template parameter.

    Example:
    ```python
    class ClassNameParameter:
        name = "class_name"
        analyse_regex = r"class (?P<name>[a-zA-Z_][a-zA-Z0-9_]*)"

        def __init__(self):
            self._class_name = None

        def analyse(self, match: dict[str, str]) -> None:
            # Store first found class name
            if self._class_name is None:
                self._class_name = match["name"]

        @property
        def regex(self) -> str:
            # After analysis, look for all occurrences of the class name
            return self._class_name

        def replace(self, match: dict[str, str]) -> str:
            # Replace with template parameter
            return f"${self.name}"

    # Usage:
    builder = TemplateBuilder()
    builder.set_prototype('''
        class MyClass:
            def method(self):
                return MyClass()
    ''')
    builder.add_parameter(ClassNameParameter())
    template = builder.build()
    print(template.render(dict(class_name="MyNewClass"))) # Will replace all occurrences of "MyClass" with "MyNewClass"
    ```
    """

    analyse_regex: str
    """Will be used during the analysis phase.
    
    See [`analyse`](#analyse) for more information.
    """

    def analyse(self, match: dict[str, str]) -> None:
        """Called for each match of `analyse_regex` during the analysis phase.

        :param match: A dictionary of the named capture groups.
        """
        ...


class TemplateBuilder:
    """Builds a template based on a given prototype.

    The builder takes a *prototype* as a string and creates a template from it,
    based on specified template parameter types.
    A template parameter type specifies how a pattern inside the prototype shall
    be converted into a template parameter.

    Two different kinds of template parameter types are supported:

    `TemplateParameterType`

    : will use a regular expression `.regex` to search for
        occurences of a pattern and replace them by the result of calling `.replace`
        on the found match. Usually the new value is a template parameter, e.g.,
        `"$my_template_param"`.
        :::{note}
        in case of multiple overlapping parameter types the first one added
        to the builder will take precedence.
        :::


    `AnalysingTemplateParameterType`

    :  works like `TemplateParameterType` but will
        first go through an analysing phase. This is useful to let the regular expression
        depend on the content of the prototype. E.g., we could find the first defined
        class in a python file and replace all occurences of that class name with
        a template parameter.

    Once build, the template is cached. The cache is invalidated as soon as
    new parameter types are added or the underlying prototype is changed.
    """

    def __init__(self) -> None:
        self._prototype = ""
        self._parameters: dict[str, _ParameterTypeWrapper] = dict()
        self._analysing_template_parameters: dict[str, _AnalyseParameterTypeWrapper] = (
            dict()
        )
        self._template = ""
        self._cached_template_is_valid = False

    def set_prototype(
        self, prototype: str | tuple[str, ...] | list[str]
    ) -> "TemplateBuilder":
        self._invalidate_cache()
        if isinstance(prototype, str):
            self._prototype = prototype
        else:
            self._prototype = "\n".join(prototype)
        return self

    def add_parameter(
        self, _type: TemplateParameter | AnalysingTemplateParameter
    ) -> "TemplateBuilder":
        self._invalidate_cache()
        self._parameter_type_adder(_type)()
        return self

    @singledispatchmethod
    def _parameter_type_adder(
        self, _type: TemplateParameter | AnalysingTemplateParameter
    ) -> Callable[[], None]:
        raise NotImplementedError()

    @_parameter_type_adder.register
    def _(self, _type: TemplateParameter) -> Callable[[], None]:
        def adder():
            name = str(len(self._parameters))
            name = "_{name}".format(name=name)
            self._parameters[name] = _ParameterTypeWrapperImpl(name, _type)

        return adder

    @_parameter_type_adder.register
    def _(self, _type: AnalysingTemplateParameter) -> Callable[[], None]:
        def adder():
            name = str(len(self._parameters))
            name = "_{name}".format(name=name)
            wrapped = _AnalyseParameterTypeWrapper(name, _type)

            self._parameters[name] = wrapped
            self._analysing_template_parameters[name] = wrapped

        return adder

    def build(self) -> str:
        if not self._cached_template_is_valid:
            self._analyse()
            regex = self._build_replacement_regex()
            self._template = regex.sub(self._replace, self._prototype)
        return self._template

    def _replace(self, m: Match) -> str:
        type_name = m.lastgroup
        if type_name is not None:
            replacement = self._parameters[type_name].replace
        else:
            raise ValueError("no match found")
        return replacement(m)

    def _analyse(self) -> None:
        regex = self._build_analyse_regex()
        for match in regex.finditer(self._prototype):
            analyse = self._get_analyser(match.lastgroup)
            analyse(match)

    def _get_analyser(self, name: str | None) -> Callable[[Match], None]:
        if name is not None and name in self._analysing_template_parameters:
            return self._analysing_template_parameters[name].analyse
        else:

            def analyse(m: Match):
                pass

            return analyse

    def _build_analyse_regex(self) -> Pattern:
        return self._build_regex(
            (name, _type.analyse_regex)
            for name, _type in self._analysing_template_parameters.items()
        )

    def _build_replacement_regex(self) -> Pattern:
        return self._build_regex(
            (name, _type.regex) for name, _type in self._parameters.items()
        )

    def _build_regex(self, types: Iterable[tuple[str, str]]) -> Pattern:
        regex = "|".join(
            "(?P<{name}>{regex})".format(name=name, regex=type_regex)
            for name, type_regex in types
        )
        return _regex_compile(regex)

    def _invalidate_cache(self) -> None:
        self._cached_template_is_valid = False


class _ParameterTypeWrapper(Protocol):
    @property
    def regex(self) -> str: ...

    @property
    def name(self) -> str: ...

    def replace(self, m: Match) -> str: ...


class _ParameterTypeWrapperImpl:
    def __init__(self, name: str, _type: TemplateParameter) -> None:
        self._type = _type
        self.name = name

    @property
    def regex(self) -> str:
        return _mangle_capture_group_names(self.name, self._type.regex)

    def replace(self, m: Match) -> str:
        return self._type.replace(
            _demangle_capture_group_names(self.name, m.groupdict())
        )


class _AnalyseParameterTypeWrapper:
    def __init__(self, name: str, _type: AnalysingTemplateParameter) -> None:
        self._type = _type
        self.name = name

    @property
    def regex(self) -> str:
        return _mangle_capture_group_names(self.name, self._type.regex)

    @property
    def analyse_regex(self) -> str:
        return _mangle_capture_group_names(self.name, self._type.analyse_regex)

    def analyse(self, match: Match) -> None:
        self._type.analyse(_demangle_capture_group_names(self.name, match.groupdict()))

    def replace(self, match: Match) -> str:
        match_dict = _demangle_capture_group_names(self.name, match.groupdict())
        return self._type.replace(match_dict)


def _mangle_capture_group_names(name, regex: str) -> str:
    capture_group_pattern = r"\(\?P<(?P<capture>[^>]+)>"

    def replace(match: Match) -> str:
        return f"(?P<{name}_{match['capture']}>"

    return _regex_compile(capture_group_pattern).sub(replace, regex)


def _demangle_capture_group_names(name, match: dict[str, str]) -> dict[str, str]:
    new_dict = {}
    for k, v in match.items():
        if k.startswith(f"{name}"):
            k = k.removeprefix(f"{name}")
            k = k.removeprefix("_")
            new_dict[k] = v
    return new_dict



from elasticai.creator import graph as gr
from elasticai.creator.graph.vf2.state import State


def test_update_in_nodes():
    state = State(graph=gr.BaseGraph().add_edge("a", "b"))
    state.add_pair("a", "first")
    state.next_depth()
    in_nodes = [state.order_back[i] for i, d in enumerate(state.in_nodes) if d == 1]
    out_nodes = [state.order_back[i] for i, d in enumerate(state.out_nodes) if d == 1]
    assert (in_nodes, out_nodes) == ([], ["b"])


def test_update_out_nodes():
    state = State(graph=gr.BaseGraph().add_edge("a", "b"))
    state.add_pair("b", "second")
    state.next_depth()
    in_nodes = [state.order_back[i] for i, d in enumerate(state.in_nodes) if d == 1]
    out_nodes = [state.order_back[i] for i, d in enumerate(state.out_nodes) if d == 1]
    assert (in_nodes, out_nodes) == (["a"], [])


def test_get_partial_succ():
    s = State(graph=gr.BaseGraph().add_edge("a", "b"))
    s.add_pair("b", "second")
    s.next_depth()
    assert s.partial_successors("a") == {"b"}


def test_get_partial_pred():
    s = State(gr.BaseGraph().add_edge("a", "b"))
    s.add_pair("a", "first")
    s.next_depth()
    assert s.partial_predecessors("b") == {"a"}


def test_get_unseen_nodes():
    s = State(gr.BaseGraph().add_edge("a", "b").add_edge("b", "c"))
    s.add_pair("a", "first")
    s.next_depth()
    assert s.unseen_nodes() == {"c"}


def test_unseen_successors():
    s = State(
        gr.BaseGraph()
        .add_edge("a", "b")
        .add_edge("b", "c")
        .add_edge("b", "d")
        .add_edge("a", "d")
    )
    s.add_pair("a", "first")
    s.next_depth()
    assert s.unseen_successors("b") == {"c"}



from typing import cast

import pytest
import torch

from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.nn.fixed_point.fxp_round_cut import (
    CutToFixedPoint,
    RoundToFixedPoint,
)
from tests.tensor_test_case import assertTensorEqual


def round_to_fxp(inputs: list[float] | torch.Tensor) -> torch.Tensor:
    if isinstance(inputs, list):
        inputs = torch.tensor(inputs)
    return cast(
        torch.Tensor,
        RoundToFixedPoint.apply(
            inputs, FxpArithmetic(FxpParams(total_bits=4, frac_bits=2, signed=True))
        ),
    )


def cut_to_fxp(inputs: list[float] | torch.Tensor) -> torch.Tensor:
    if isinstance(inputs, list):
        inputs = torch.tensor(inputs)
    return cast(
        torch.Tensor,
        CutToFixedPoint.apply(
            inputs, FxpArithmetic(FxpParams(total_bits=4, frac_bits=2, signed=True))
        ),
    )


@pytest.mark.parametrize(
    "val_in, val_out",
    [(1.75, 1.75), (-2.0, -2.0), (-1.3, -1.25), (0.1, 0.0), (1.6, 1.5)],
)
def test_round_integer(val_in: float, val_out: float) -> None:
    assertTensorEqual(
        expected=round_to_fxp([val_in]),
        actual=[val_out],
    )


@pytest.mark.parametrize("val_in, val_out", [(-2.5, -2.0), (2.0, 1.75)])
def test_round_out_of_bounds(val_in: float, val_out: float) -> None:
    result = torch.tensor([val_in])
    with pytest.raises(ValueError):
        _ = round_to_fxp(result)


@pytest.mark.parametrize(
    "val_in, val_out",
    [(1.75, 1.75), (-2.0, -2.0), (-1.3, -1.25), (0.1, 0.0), (1.6, 1.5)],
)
def test_cut_integer(val_in: float, val_out: float) -> None:
    assertTensorEqual(
        expected=cut_to_fxp([val_in]),
        actual=[val_out],
    )


@pytest.mark.parametrize("val_in, val_out", [(-2.5, -2.0), (2.0, 1.75)])
def test_cut_out_of_bounds(val_in: float, val_out: float) -> None:
    inputs = torch.tensor([val_in])
    with pytest.raises(ValueError):
        _ = cut_to_fxp(inputs)



from typing import cast

import pytest

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.fixed_point import BatchNormedLinear


@pytest.fixture
def batchnorm_linear_design() -> BatchNormedLinear:
    return BatchNormedLinear(
        total_bits=16,
        frac_bits=8,
        in_features=3,
        out_features=2,
    )


def save_design(design: BatchNormedLinear) -> dict[str, str]:
    destination = InMemoryPath("batchnorm_linear", parent=None)
    design.create_design("batchnorm_linear").save_to(destination)
    files = cast(list[InMemoryFile], list(destination.children.values()))
    return {file.name: "\n".join(file.text) for file in files}


def test_saved_design_contains_needed_files(
    batchnorm_linear_design: BatchNormedLinear,
) -> None:
    saved_files = save_design(batchnorm_linear_design)
    expected_files = {
        "batchnorm_linear_rom.vhd",
        "batchnorm_linear.vhd",
    }
    actual_files = set(saved_files.keys())
    assert expected_files == actual_files



import unittest

from elasticai.creator_plugins.grouped_filter import FilterParameters
from elasticai.creator_plugins.grouped_filter.src.index_generators import (
    GroupedFilterIndexGenerator,
)


def to_tuple(xs):
    if hasattr(xs, "__iter__"):
        return tuple(to_tuple(x) for x in xs)
    return xs


class UnrolledFilterParameterIndicesTest(unittest.TestCase):
    def test_respects_channelwise_interleaving_for_steps(self):
        expected = (((0, 1),), ((2, 3),))
        actual = GroupedFilterIndexGenerator(
            params=FilterParameters(
                input_size=2,
                in_channels=2,
                kernel_size=1,
                groups=1,
                stride=1,
                out_channels=1,
            ),
        ).as_tuple_by_steps()
        self.assertEqual(expected, actual)

    def test_using_two_groups_for_two_channels_separates_channels(self):
        expected = (((0,), (1,)), ((2,), (3,)))
        actual = GroupedFilterIndexGenerator(
            params=FilterParameters(
                input_size=2,
                in_channels=2,
                kernel_size=1,
                groups=2,
                stride=1,
                out_channels=2,
            ),
        ).as_tuple_by_steps()
        self.assertEqual(expected, actual)

    def test_generate_3_steps_for_spatial_size_5_and_window_size_3(self):
        indices = GroupedFilterIndexGenerator(
            params=FilterParameters(
                input_size=5,
                kernel_size=3,
                in_channels=1,
                groups=1,
                stride=1,
                out_channels=1,
            ),
        ).as_tuple_by_steps()
        self.assertEqual(3, len(indices))

    def test_time_steps_from_same_channel_are_concatenated_across_window(self):
        expected = (
            (
                (
                    0,
                    2,
                ),
                (
                    1,
                    3,
                ),
            ),
        )
        indices = GroupedFilterIndexGenerator(
            params=FilterParameters(
                input_size=2,
                kernel_size=2,
                in_channels=2,
                groups=2,
                stride=1,
                out_channels=2,
            ),
        ).as_tuple_by_steps()
        self.assertEqual(expected, indices)

    def test_strides_are_correctly_concatenated(self):
        expected = (((0, 1, 4, 5), (2, 3, 6, 7)), ((4, 5, 8, 9), (6, 7, 10, 11)))
        generator = GroupedFilterIndexGenerator(
            params=FilterParameters(
                input_size=3,
                kernel_size=2,
                in_channels=4,
                groups=2,
                stride=1,
                out_channels=2,
            ),
        )
        indices = generator.as_tuple_by_steps()

        self.assertEqual(expected, indices)

    def test_index_generator_interface(self):
        expected = [[[0, 1, 4, 5], [2, 3, 6, 7]], [[4, 5, 8, 9], [6, 7, 10, 11]]]
        generator = GroupedFilterIndexGenerator(
            params=FilterParameters(
                input_size=3,
                kernel_size=2,
                in_channels=4,
                groups=2,
                stride=1,
                out_channels=2,
            ),
        )
        actual = []
        for step in generator.steps():
            groups = []
            for group in step.groups():
                indices = list(group)
                groups.append(indices)
            actual.append(groups)

        self.assertEqual(expected, actual)

    def test_index_generator_interface_by_groups(self):
        expected = [
            [[0, 1, 4, 5], [4, 5, 8, 9]],
            [[2, 3, 6, 7], [6, 7, 10, 11]],
        ]
        generator = GroupedFilterIndexGenerator(
            params=FilterParameters(
                input_size=3,
                kernel_size=2,
                in_channels=4,
                groups=2,
                stride=1,
                out_channels=2,
            ),
        )
        actual = []
        for group in generator.groups():
            steps = []
            for step in group.steps():
                indices = list(step)
                steps.append(indices)
            actual.append(steps)

        self.assertEqual(expected, actual)



from typing import Callable

from torch import Tensor
from torch.nn import Module

from .quantize_to_fixed_point import quantize_to_fxp_hte, quantize_to_fxp_stochastic
from .two_complement_fixed_point_config import FixedPointConfigV2


def get_quantize_to_fixed_point(
    func: Callable[[Tensor, Tensor, Tensor, Tensor], Tensor],
) -> tuple[type[Module], type[Module]]:
    class QuantizeToFixedPoint(Module):
        def __init__(self, config: FixedPointConfigV2):
            super().__init__()
            self.register_buffer("minimum_as_rational", config.minimum_as_rational)
            self.register_buffer("maximum_as_rational", config.maximum_as_rational)
            self.register_buffer("resolution_per_int", config.resolution_per_int)

        @staticmethod
        def forward(x: Tensor) -> Tensor:
            return x

        def right_inverse(self, x: Tensor) -> Tensor:
            return func(
                x,
                self.resolution_per_int,
                self.minimum_as_rational,
                self.maximum_as_rational,
            )

    class QuantizeToFixedPointSTE(Module):
        def __init__(self, config: FixedPointConfigV2):
            super().__init__()
            self.register_buffer("minimum_as_rational", config.minimum_as_rational)
            self.register_buffer("maximum_as_rational", config.maximum_as_rational)
            self.register_buffer("resolution_per_int", config.resolution_per_int)

        def forward(self, x: Tensor) -> Tensor:
            return func(
                x,
                self.resolution_per_int,
                self.minimum_as_rational,
                self.maximum_as_rational,
            )

        def right_inverse(self, x: Tensor) -> Tensor:
            return x

    return QuantizeToFixedPoint, QuantizeToFixedPointSTE


(QuantizeParamToFixedPointHTE, QuantizeParamSTEToFixedPointHTE) = (
    get_quantize_to_fixed_point(quantize_to_fxp_hte)
)
(QuantizeParamToFixedPointStochastic, QuantizeParamSTEToFixedPointStochastic) = (
    get_quantize_to_fixed_point(quantize_to_fxp_stochastic)
)


class QuantizeTensorToFixedPointHTE(QuantizeParamSTEToFixedPointHTE):
    """
    This Modules can be used for Tensor quantization
    """


class QuantizeTensorToFixedPointStochastic(QuantizeParamSTEToFixedPointStochastic):
    """
    This Modules can be used for Tensor quantization
    """



import logging

import pytest

import elasticai.creator.graph as gr
from elasticai.creator.graph.graph_rewriting import RewriteResult
from tests.unit_tests.graph.utils import (
    Graph,
    get_rewriter,
    get_rewriter_returning_full_result,
)

from .utils import (
    build_graph_from_dict,
)

"""Tests
- fail if removing pattern match leaves dangling edges
"""


def test_can_replace_single_node():
    graph = build_graph_from_dict(
        {("a", "a_t"): ["b"], ("b", "b_t"): ["c"], ("c", "c_t"): []}
    )
    pattern = build_graph_from_dict(
        {("0", "a_t"): ["1"], ("1", "b_t"): ["2"], ("2", "c_t"): []}
    )

    replacement = build_graph_from_dict(
        {("r0", "a_t"): ["r1"], ("r1", "r_t"): ["r2"], ("r2", "c_t"): []}
    )

    lhs = {"i0": "0", "i1": "2"}

    rhs = {"i0": "r0", "i1": "r2"}

    rewrite = get_rewriter(pattern, replacement, lhs, rhs)

    expected = build_graph_from_dict(
        {
            ("c", "c_t"): [],
            ("a", "a_t"): ["r1"],
            ("r1", "r_t"): ["c"],
        }
    )
    assert rewrite(graph).as_dict() == expected.as_dict()


def test_create_new_name_for_replacement_in_case_of_conflict():
    graph = build_graph_from_dict(
        {("a", "a_t"): ["b"], ("b", "b_t"): ["c"], ("c", "c_t"): []}
    )
    pattern = build_graph_from_dict(
        {("0", "a_t"): ["1"], ("1", "b_t"): ["2"], ("2", "c_t"): []}
    )

    replacement = build_graph_from_dict(
        {("r0", "a_t"): ["a"], ("a", "r_t"): ["r2"], ("r2", "c_t"): []}
    )

    lhs = {"i0": "0", "i1": "2"}

    rhs = {"i0": "r0", "i1": "r2"}

    rewrite = get_rewriter(pattern, replacement, lhs, rhs)

    expected = build_graph_from_dict(
        {
            ("c", "c_t"): [],
            ("a", "a_t"): ["a_1"],
            ("a_1", "r_t"): ["c"],
        }
    )
    assert rewrite(graph).as_dict() == expected.as_dict()


def test_raise_error_if_rhs_is_not_injective():
    pattern = build_graph_from_dict(
        {("0", "a_t"): ["1"], ("1", "b_t"): ["2"], ("2", "c_t"): []}
    )

    replacement = build_graph_from_dict(
        {("r0", "a_t"): ["a"], ("a", "r_t"): ["r2"], ("r2", "c_t"): []}
    )

    lhs = {"i0": "0", "i1": "2"}
    rhs = {"i0": "r0", "i1": "r0"}
    rewriter = get_rewriter(pattern, replacement, lhs, rhs)
    pytest.raises(
        ValueError,
        rewriter,
        build_graph_from_dict(
            {("a", "a_t"): ["b"], ("b", "b_t"): ["c"], ("c", "c_t"): []}
        ),
    )


def test_can_replace_one_of_two_matches():
    graph = build_graph_from_dict(
        {
            ("a", "a_t"): ["b"],
            ("b", "b_t"): ["c"],
            ("c", "a_t"): ["d"],
            ("d", "b_t"): ["e"],
            ("e", "a_t"): [],
        }
    )
    pattern = build_graph_from_dict(
        {("0", "a_t"): ["1"], ("1", "b_t"): ["2"], ("2", "a_t"): []}
    )

    replacement = build_graph_from_dict(
        {("r0", "a_t"): ["r1"], ("r1", "r_t"): ["r2"], ("r2", "c_t"): []}
    )

    lhs = {"i0": "0", "i1": "2"}
    rhs = {"i0": "r0", "i1": "r2"}

    rewrite = get_rewriter(pattern, replacement, lhs, rhs)

    expected = build_graph_from_dict(
        {
            ("c", "a_t"): ["d"],
            ("a", "a_t"): ["r1"],
            ("r1", "r_t"): ["c"],
            ("d", "b_t"): ["e"],
            ("e", "a_t"): [],
        }
    )
    actual = rewrite(graph)
    assert set(actual.edges) == set(expected.edges)
    assert set(actual.nodes) == set(expected.nodes)


def test_raise_error_if_rewrite_would_produce_dangling_edge():
    g = (
        gr.BaseGraph()
        .add_edge("a", "b")
        .add_edge("b", "c")
        .add_edge("c", "d")
        .add_edge("would_be_disconnected", "b")
    )
    raised = False
    try:
        gr.rewrite(
            replacement=gr.BaseGraph().add_edge("i0", "r").add_edge("r", "i1"),
            original=g,
            match={"i0": "a", "p0": "b", "p1": "c", "i1": "d"},
            lhs={"i0": "i0", "i1": "i1"},
            rhs={"i0": "i0", "i1": "i1"},
        )
    except gr.DanglingEdgeError:
        raised = True

    assert raised


@pytest.fixture
def graph_for_multiple_matches():
    graph = gr.BaseGraph()
    sequence = ["a", "b", "c", "d", "e", "f", "g", "h"]
    for e in zip(sequence[:-1], sequence[1:]):
        graph.add_edge(e[0], e[1])
    return graph


def test_replacing_all_matches_fails(graph_for_multiple_matches):
    sequence = ("a", "b", "c", "d", "e", "f", "g", "h")
    all_matches = list(zip(sequence[:-2], sequence[1:-1], sequence[2:]))
    all_matches = [{"0": x, "1": y, "2": z} for x, y, z in all_matches]
    replacement = gr.BaseGraph().add_edge("r0", "r1").add_edge("r1", "r2")
    lhs = {"i0": "0", "i1": "2"}
    rhs = {"i0": "r0", "i1": "r2"}
    result = graph_for_multiple_matches
    raised = False
    try:
        for i, match in enumerate(all_matches):
            result, replacement_map = gr.rewrite(
                replacement=replacement,
                original=result,
                match=match,
                lhs=lhs,
                rhs=rhs,
            )
    except gr.DanglingEdgeError:
        raised = True

    assert raised and i == 1, (
        "Expected to raise DanglingEdgeError when rewriting second match, but did not."
    )


def test_can_replace_multiple_matches(graph_for_multiple_matches):
    graph = graph_for_multiple_matches
    safe_matches = [("a", "b", "c"), ("c", "d", "e"), ("e", "f", "g")]
    safe_matches = [{"0": x, "1": y, "2": z} for x, y, z in safe_matches]
    replacement = gr.BaseGraph().add_edge("r0", "r1").add_edge("r1", "r2")
    result = graph
    for match in safe_matches:
        result, replacement_map = gr.rewrite(
            replacement=replacement,
            original=result,
            match=match,
            lhs={"i0": "0", "i1": "2"},
            rhs={"i0": "r0", "i1": "r2"},
        )
    expected_sequence = ["a", "r1", "c", "r1_1", "e", "r1_2", "g", "h"]
    expected_edges = set(zip(expected_sequence[:-1], expected_sequence[1:]))
    assert set(result.iter_edges()) == expected_edges


class TestRewriteResult:
    @pytest.fixture
    def graph(self):
        return build_graph_from_dict(
            {("a", "a_t"): ["b"], ("b", "b_t"): ["c"], ("c", "c_t"): []}
        )

    @pytest.fixture
    def pattern(self):
        return build_graph_from_dict(
            {("0", "a_t"): ["1"], ("1", "b_t"): ["2"], ("2", "c_t"): []}
        )

    @pytest.fixture
    def replacement(self):
        return build_graph_from_dict(
            {("r0", "a_t"): ["r1"], ("r1", "r_t"): ["r2"], ("r2", "c_t"): []}
        )

    @pytest.fixture
    def rhs_dict(self):
        return {"i0": "r0", "i1": "r2"}

    @pytest.fixture
    def result(
        self,
        graph: Graph,
        pattern: Graph,
        replacement: Graph,
        rhs_dict: dict[str, str],
    ):
        lhs = {"i0": "0", "i1": "2"}

        rhs = rhs_dict

        rewrite = get_rewriter_returning_full_result(pattern, replacement, lhs, rhs)
        result = rewrite(graph)
        return result

    def test_can_map_pattern_back_to_original_graph(self, result: RewriteResult):
        expected_pattern_to_original = {"0": "a", "1": "b", "2": "c"}

        assert result.pattern_to_original == expected_pattern_to_original

    def test_can_map_replacement_to_new_graph(self, result: RewriteResult):
        expected_replacement_to_new = {"r0": "a", "r1": "r1", "r2": "c"}

        assert result.replacement_to_new == expected_replacement_to_new

    def test_can_combine_attribute_from_replacement_and_original_graph(
        self,
        result: RewriteResult,
        graph: Graph,
        pattern: Graph,
        replacement: Graph,
        rhs_dict: dict[str, str],
    ):
        graph.data = {"a": "1", "b": "2", "c": "3"}
        replacement.data = {"r0": "100", "r1": "200", "r2": "300"}

        new_graph = Graph(result.new_graph, {})
        for repl_node in replacement.data:
            new_node = result.replacement_to_new[repl_node]
            original_node_with_interesting_value = result.pattern_to_original["1"]
            interesting_value = graph.data[original_node_with_interesting_value]
            replacement_value = replacement.data[repl_node]
            not_in_interface = repl_node not in rhs_dict.values()
            if not_in_interface:
                new_graph.data[new_node] = str(
                    int(replacement_value) + int(interesting_value)
                )
            else:
                new_graph.data[new_node] = graph.data[new_node]

        assert new_graph.data == {"a": "1", "r1": "202", "c": "3"}


def create_real_world_graph_with_data() -> tuple[
    gr.Graph[str], dict[str, dict[str, str]]
]:
    data = {
        "input_1": {
            "type": "input",
        },
        "conv0_0": {
            "type": "conv1d",
        },
        "conv0_1": {
            "type": "maxpool1d",
        },
        "conv0_2": {
            "type": "batchnorm1d",
        },
        "conv0_3": {
            "type": "binarize",
        },
        "conv1_0": {
            "type": "conv1d",
        },
        "conv1_1": {
            "type": "maxpool1d",
        },
        "conv1_2": {
            "type": "batchnorm1d",
        },
        "conv1_3": {
            "type": "binarize",
        },
        "conv2_0": {
            "type": "conv1d",
        },
        "conv2_1": {
            "type": "maxpool1d",
        },
        "conv2_2": {
            "type": "batchnorm1d",
        },
        "conv2_3": {
            "type": "binarize",
        },
        "conv3_0": {
            "type": "conv1d",
        },
        "conv3_1": {
            "type": "maxpool1d",
        },
        "conv3_2": {
            "type": "batchnorm1d",
        },
        "conv3_3": {
            "type": "binarize",
        },
        "conv4_0": {
            "type": "conv1d",
        },
        "conv4_1": {
            "type": "maxpool1d",
        },
        "conv4_2": {
            "type": "batchnorm1d",
        },
        "conv4_3": {
            "type": "binarize",
        },
        "flatten": {
            "type": "flatten",
        },
        "lin": {
            "type": "linear",
        },
        "sigmoid": {
            "type": "sigmoid",
        },
        "flatten1": {
            "type": "flatten",
        },
        "output": {
            "type": "output",
        },
    }
    g: gr.Graph[str] = gr.BaseGraph()
    original_edges = {
        "input_1": {"conv0_0"},
        "conv0_0": {"conv0_1"},
        "conv0_1": {"conv0_2"},
        "conv0_2": {"conv0_3"},
        "conv0_3": {"conv1_0"},
        "conv1_0": {"conv1_1"},
        "conv1_1": {"conv1_2"},
        "conv1_2": {"conv1_3"},
        "conv1_3": {"conv2_0"},
        "conv2_0": {"conv2_1"},
        "conv2_1": {"conv2_2"},
        "conv2_2": {"conv2_3"},
        "conv2_3": {"conv3_0"},
        "conv3_0": {"conv3_1"},
        "conv3_1": {"conv3_2"},
        "conv3_2": {"conv3_3"},
        "conv3_3": {"conv4_0"},
        "conv4_0": {"conv4_1"},
        "conv4_1": {"conv4_2"},
        "conv4_2": {"conv4_3"},
        "conv4_3": {"flatten"},
        "flatten": {"lin"},
        "lin": {"sigmoid"},
        "sigmoid": {"flatten1"},
        "flatten1": {"output"},
    }
    for src in original_edges:
        for dst in original_edges[src]:
            g.add_edge(src, dst)

    return g, data


@pytest.fixture
def expected_edges_after_complex_reordering():
    expected_edges = {
        "input_1": {"conv0_0"},
        "conv0_0": {"batchnorm1d"},
        "batchnorm1d": {"binarize"},
        "binarize": {"maxpool1d"},
        "maxpool1d": {"conv1_0"},
        "conv1_0": {"batchnorm1d_1"},
        "batchnorm1d_1": {"binarize_1"},
        "binarize_1": {"maxpool1d_1"},
        "maxpool1d_1": {"conv2_0"},
        "conv2_0": {"batchnorm1d_2"},
        "batchnorm1d_2": {"binarize_2"},
        "binarize_2": {"maxpool1d_2"},
        "maxpool1d_2": {"conv3_0"},
        "conv3_0": {"batchnorm1d_3"},
        "batchnorm1d_3": {"binarize_3"},
        "binarize_3": {"maxpool1d_3"},
        "maxpool1d_3": {"conv4_0"},
        "conv4_0": {"batchnorm1d_4"},
        "batchnorm1d_4": {"binarize_4"},
        "binarize_4": {"maxpool1d_4"},
        "maxpool1d_4": {"flatten"},
        "flatten": {"lin"},
        "lin": {"sigmoid"},
        "sigmoid": {"flatten1"},
        "flatten1": {"output"},
        "output": set(),
    }
    return expected_edges


def test_can_reorder_complex_layer_sequence(expected_edges_after_complex_reordering):
    logger = logging.getLogger("elasticai.creator.graph")

    def make_pattern(data):
        """Also consider the corresponding constraint for this pattern.

        The pattern is defined in such a way that we declare potentially
        matched convolutional layers to be part of the interface:
        'start', 'end'. Thus, we "promise" that we will not edit
        convolutional layers. Instead, we will only reorder
        maxpool1d, batchnorm1d, and binarize layers.
        """
        pattern = gr.BaseGraph()
        pattern_sequence = [
            "start",
            "maxpool1d",
            "batchnorm1d",
            "binarize",
            "end",
        ]

        for src, dst in zip(pattern_sequence[:-1], pattern_sequence[1:]):
            pattern.add_edge(src, dst)

        def constraint(pattern_node: str, graph_node: str) -> bool:
            match pattern_node:
                case "start":
                    return data[graph_node]["type"] in ("input", "conv1d")
                case "end":
                    return data[graph_node]["type"] in ("output", "conv1d", "flatten")
                case _:
                    return data[graph_node]["type"] == pattern_node

        return pattern, constraint

    def make_replacement():
        replacement_sequence = [
            "start",
            "batchnorm1d",
            "binarize",
            "maxpool1d",
            "end",
        ]
        replacement = gr.BaseGraph()

        for src, dst in zip(replacement_sequence[:-1], replacement_sequence[1:]):
            replacement.add_edge(src, dst)
        return replacement

    g, data = create_real_world_graph_with_data()
    pattern, constraint = make_pattern(data)
    matches = list(
        gr.find_all_subgraphs(
            pattern=pattern,
            graph=g,
            node_constraint=constraint,
        )
    )

    def log(msg):
        logger.log(logging.DEBUG, msg)

    log(f"Found {len(matches)} matches")
    for match in matches:
        log(f"Match: {match}")

    lhs = {"i0": "start", "i1": "end"}
    safe_matches = list(
        gr.get_rewriteable_matches(g, matches, interface_nodes=lhs.values())
    )

    log(f"Safe matches: {len(safe_matches)}")
    for match in safe_matches:
        log(f"Match: {match}")

    def rewrite_all_matches(g, matches, replacement, lhs, rhs):
        result = g
        for match in matches:
            result, replacement_map = gr.rewrite(
                replacement=replacement, original=result, match=match, lhs=lhs, rhs=rhs
            )
        return result

    def visit_nodes(graph, node, visited):
        if node not in visited:
            visited.add(node)
            log(node)
            for successor in graph.successors[node]:
                visit_nodes(graph, successor, visited)

    replacement = make_replacement()
    rhs = lhs
    result = rewrite_all_matches(g, safe_matches, replacement, lhs, rhs)

    visit_nodes(result, "input_1", set())
    assert expected_edges_after_complex_reordering == dict(result.successors)



from typing import cast

import pytest

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.identity.layer import BufferedIdentity
from elasticai.creator.nn.sequential.layer import Sequential


class TestSequential:
    def test_empty_sequential_passes_through_signals(self) -> None:
        expected = """library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.all;

entity sequential is
    port (
        enable: in std_logic;
        clock: in std_logic;

        x_address: out std_logic_vector(1-1 downto 0);
        y_address: in std_logic_vector(1-1 downto 0);

        x: in std_logic_vector(1-1 downto 0);
        y: out std_logic_vector(1-1 downto 0);

        done: out std_logic
    );
end sequential;

architecture rtl of sequential is
begin
    done <= enable;
    x_address <= y_address;
    y <= x;
    --------------------------------------------------------------------------------
    -- Instantiate all layers
    --------------------------------------------------------------------------------
end rtl;"""
        actual_code = "\n".join(sequential_layer_code_for_model(Sequential()))
        assert actual_code == expected

    @pytest.mark.parametrize(
        "number_of_layer_inputs, address_width", [(2, 1), (3, 2), (8, 3)]
    )
    def test_single_layer_model(self, number_of_layer_inputs, address_width) -> None:
        expected = f"""library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.all;

entity sequential is
    port (
        enable: in std_logic;
        clock: in std_logic;

        x_address: out std_logic_vector({address_width}-1 downto 0);
        y_address: in std_logic_vector({address_width}-1 downto 0);

        x: in std_logic_vector(4-1 downto 0);
        y: out std_logic_vector(4-1 downto 0);

        done: out std_logic
    );
end sequential;

architecture rtl of sequential is
    signal i_bufferedidentity_0_clock : std_logic := '0';
    signal i_bufferedidentity_0_done : std_logic := '0';
    signal i_bufferedidentity_0_enable : std_logic := '0';
    signal i_bufferedidentity_0_x : std_logic_vector(3 downto 0) := (others => '0');
    signal i_bufferedidentity_0_x_address : std_logic_vector({address_width - 1} downto 0) := (others => '0');
    signal i_bufferedidentity_0_y : std_logic_vector(3 downto 0) := (others => '0');
    signal i_bufferedidentity_0_y_address : std_logic_vector({address_width - 1} downto 0) := (others => '0');
begin
    done <= i_bufferedidentity_0_done;
    i_bufferedidentity_0_clock <= clock;
    i_bufferedidentity_0_enable <= enable;
    i_bufferedidentity_0_x <= x;
    i_bufferedidentity_0_y_address <= y_address;
    x_address <= i_bufferedidentity_0_x_address;
    y <= i_bufferedidentity_0_y;
    --------------------------------------------------------------------------------
    -- Instantiate all layers
    --------------------------------------------------------------------------------
    i_bufferedidentity_0 : entity work.bufferedidentity_0(rtl)
    port map(
      clock => i_bufferedidentity_0_clock,
      done => i_bufferedidentity_0_done,
      enable => i_bufferedidentity_0_enable,
      x => i_bufferedidentity_0_x,
      x_address => i_bufferedidentity_0_x_address,
      y => i_bufferedidentity_0_y,
      y_address => i_bufferedidentity_0_y_address
    );
end rtl;"""
        model = Sequential(
            BufferedIdentity(num_input_features=number_of_layer_inputs, total_bits=4)
        )
        actual_code = "\n".join(sequential_layer_code_for_model(model))
        assert actual_code == expected


def get_code(code_file: InMemoryPath | InMemoryFile) -> list[str]:
    return cast(InMemoryFile, code_file).text


def translate_model(model: Sequential) -> InMemoryPath:
    design = model.create_design("sequential")
    destination = InMemoryPath("sequential", parent=None)
    design.save_to(destination)
    return destination


def sequential_layer_code_for_model(model: Sequential) -> list[str]:
    destination = translate_model(model)
    return get_code(destination["sequential"])



from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.nn.fixed_point.hard_sigmoid import HardSigmoid
from elasticai.creator.nn.fixed_point.hard_tanh.design import HardTanh
from elasticai.creator.nn.fixed_point.linear.design import LinearDesign as FPLinear1d
from elasticai.creator.vhdl.code_generation.addressable import calculate_address_width
from elasticai.creator.vhdl.design import std_signals
from elasticai.creator.vhdl.design.design import Design
from elasticai.creator.vhdl.design.ports import Port
from elasticai.creator.vhdl.design.signal import Signal
from elasticai.creator.vhdl.shared_designs.rom import Rom

__all__ = [
    "Path",
    "Design",
    "InProjectTemplate",
    "Path",
    "Port",
    "Rom",
    "Signal",
    "calculate_address_width",
    "module_to_package",
    "std_signals",
    "HardSigmoid",
    "HardTanh",
    "FPLinear1d",
]



import warnings

from elasticai.creator.arithmetic import FxpConverter, FxpParams
from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.design.ports import Port


class Skeleton:
    def __init__(
        self,
        x_num_values: int,
        y_num_values: int,
        network_name: str,
        port: Port,
        id: list[int] | int,
        skeleton_version: str = "v1",
    ):
        self.name = "skeleton"
        self._network_name = network_name
        self._port = port
        self._x_num_values = str(x_num_values)
        self._y_num_values = str(y_num_values)
        if isinstance(id, int):
            id = [id]
        self._id = id
        if skeleton_version == "v1":
            warnings.warn(
                (
                    "Skeleton V1 might be deprecated in the future. Consider using"
                    " Skeleton V2 instead."
                ),
                FutureWarning,
            )
            self._template_file_name = "network_skeleton.tpl.vhd"
            if len(id) != 1:
                raise Exception(
                    f"should give an id of 1 byte. Actual length is {len(id)}"
                )
            if x_num_values > 100:
                raise Exception(
                    "Not more than 100 input values allowed. Actual num of inputs"
                    f" {x_num_values} ."
                )
            if y_num_values > 100:
                raise Exception(
                    "Not more than 100 input values allowed. Actual num of inputs"
                    f" {x_num_values} ."
                )
        elif skeleton_version == "v2":
            self._template_file_name = "network_skeleton_v2.tpl.vhd"
            if len(id) != 16:
                raise Exception(
                    f"should give an id of 16 byte. Actual length is {len(id)}"
                )
            if x_num_values > 19983:
                raise Exception(
                    "Not more than 19983 input values allowed. Actual num of inputs"
                    f" {x_num_values} ."
                )
            if y_num_values > 19983:
                raise Exception(
                    "Not more than 19983 input values allowed. Actual num of inputs"
                    f" {x_num_values} ."
                )
        else:
            raise Exception(f"Skeleton version {skeleton_version} does not exist")
        if port["x"].width > 8:
            raise Exception(
                "port x width should not be bigger than 8. You assigned "
                f" {port['x'].width=}"
            )
        if port["y"].width > 8:
            raise Exception(
                "port x width should not be bigger than 8. You assigned "
                f" {port['y'].width=}"
            )

    def save_to(self, destination: Path):
        conv = FxpConverter(FxpParams(total_bits=8, frac_bits=0, signed=False))
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name=self._template_file_name,
            parameters=dict(
                name=self.name,
                network_name=self._network_name,
                data_width_in=str(self._port["x"].width),
                x_addr_width=str(self._port["x_address"].width),
                x_num_values=self._x_num_values,
                y_num_values=self._y_num_values,
                data_width_out=str(self._port["y"].width),
                y_addr_width=str(self._port["y_address"].width),
                id=", ".join(map(conv.integer_to_binary_string_vhdl, self._id)),
            ),
        )
        file = destination.as_file(".vhd")
        file.write(template)


class LSTMSkeleton:
    def __init__(self, network_name: str):
        self.name = "skeleton"
        self._network_name = network_name

    def save_to(self, destination: Path):
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="lstm_network_skeleton.tpl.vhd",
            parameters=dict(name=self.name, network_name=self._network_name),
        )
        file = destination.as_file(".vhd")
        file.write(template)


class EchoSkeletonV2:
    def __init__(self, num_values: int, bitwidth: int):
        self._num_values = num_values
        self._bitwidth = bitwidth
        self._name = "skeleton"
        if bitwidth > 8:
            raise Exception(
                "Not more than 8 bit supported by middleware. You assigned"
                f" {bitwidth} bits"
            )
        self._template_file_name = "network_skeleton_v2_echo.tpl.vhd"
        self._id = [50, 52, 48, 56, 50, 51, 69, 67, 72, 79, 83, 69, 82, 86, 69, 82]

    def save_to(self, destination: Path):
        conv = FxpConverter(FxpParams(total_bits=8, frac_bits=0, signed=False))
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name=self._template_file_name,
            parameters=dict(
                name=self._name,
                data_width=str(self._bitwidth),
                num_values=str(self._num_values),
                id=", ".join(map(conv.integer_to_binary_string_vhdl, self._id)),
            ),
        )
        file = destination.as_file(".vhd")
        file.write(template)



import importlib.resources as res
from string import Template

from elasticai.creator.ir2vhdl import (
    Code,
    EntityTemplateDirector,
    Implementation,
    Ir2Vhdl,
    PluginSymbol,
)

from ._skeleton import skeleton as skeleton


class _BufferedNetworkWrapper(PluginSymbol):
    def __init__(self, template: Template):
        self._template = template

    def __call__(self, args: Implementation) -> Code:
        def _iter():
            yield self._template.substitute(
                dict(entity=args.name) | args.attributes["generic_map"]
            )

        return args.name, _iter()

    @classmethod
    def load_into(cls, receiver: Ir2Vhdl) -> None:
        def load_file():
            return (
                res.files("elasticai.creator_plugins.skeleton")
                .joinpath("vhdl/buffered_network_wrapper.vhd")
                .read_text()
            )

        template = (
            EntityTemplateDirector()
            .set_prototype(load_file())
            .add_generic("KERNEL_SIZE")
            .add_generic("STRIDE")
        ).build()
        receiver.register("buffered_network_wrapper", cls(template))


buffered_network_wrapper = _BufferedNetworkWrapper



from string import Template as _pyTemplate
from typing import Mapping, Self, cast

from elasticai.creator import template as tpl


class _Definition(tpl.TemplateParameter):
    def __init__(self, type: str, name: str, delimiter: str) -> None:
        self.name = name
        self.delimiter = delimiter
        self.regex = r"(?P<param>{type}\s+(signed\s+)?(\[.*?\]\s+)?{name}\s*=)\s?.([^;,\n]|,(?=.*}}))*".format(
            type=type, name=name
        )

    def replace(self, m: dict[str, str]) -> str:
        d = self.delimiter
        return f"{m['param']} {d}{self.name}"


class _LocalParameter(_Definition):
    """Turn the localparam `name` into a template parameter."""

    def __init__(self, name: str, delimiter: str):
        super().__init__("localparam", name, delimiter)


class _Parameter(_Definition):
    def __init__(self, name: str, delimiter: str):
        super().__init__("parameter", name, delimiter)


class _IdParameter(tpl.TemplateParameter):
    def __init__(self, name: str, delimiter: str):
        self.name = name
        self.delimiter = delimiter
        self.regex = r"(?P<prefix>[^_a-zA-Z0-9]*)(?P<id>{name})(?P<suffix>[^_a-zA-Z0-9]*)".format(
            name=name
        )

    def replace(self, m: dict[str, str]) -> str:
        d = self.delimiter
        return f"{m['prefix']}{d}{self.name}{m['suffix']}"


class _ModuleOfInstance(tpl.TemplateParameter):
    def __init__(self, name: str, delimiter: str):
        self.name = name
        self.delimiter = delimiter
        self.regex = (
            r"\s*(?P<prefix>{name}(\s*#\(.*\))? )[a-zA-Z0-9_]+(?=\s*\()".format(
                name=name
            )
        )

    def replace(self, m: dict[str, str]) -> str:
        d = self.delimiter
        return f"{m['prefix']}{d}{self.name}"


class _InstanceName(tpl.TemplateParameter):
    def __init__(self, name: str, delimiter: str):
        self.name = name
        self.delimiter = delimiter
        self.regex = (
            r"\s*{name}(?P<prefix>(\s*#\(.*\))? [a-zA-Z0-9_]+(?=\s*\())".format(
                name=name
            )
        )

    def replace(self, m: dict[str, str]) -> str:
        d = self.delimiter
        pos = m[""].find(self.name)
        chck = m[""][:pos] + d + m[""][pos:]
        return chck


class _DefineSwitch(tpl.TemplateParameter):
    def __init__(self, name: str, delimiter: str):
        self.name = name
        self.regex = r"(//+)?\s*`define\s+{name}(?=\s)?".format(name=name)
        self.delimiter = delimiter

    def replace(self, m: dict[str, str]) -> str:
        d = self.delimiter
        return f"{d}def{self.name}"


class _ModuleName(tpl.AnalysingTemplateParameter):
    def __init__(self, delimiter: str):
        self._pre_analysis_regex = (
            r"\s*(?P<prefix>[^_a-zA-Z0-9]){name}(?P<suffix>[^_a-zA-Z0-9])"
        )
        self.regex = ""
        self.analyse_regex = r"\s*module\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)"
        self._analysed = False
        self._delimiter = delimiter

    def analyse(self, match: dict[str, str]) -> None:
        self._analysed = True
        name = match["name"]
        self.regex = self._pre_analysis_regex.format(name=name)

    def replace(self, match: dict[str, str]) -> str:
        d = self._delimiter
        if not self._analysed:
            raise ValueError("No module name found")
        return f"{match['prefix']}{d}module_name{match['suffix']}"


class TemplateDirector:
    """Director for building verilog templates.

    Most methods correspond to verilog language constructs.
    Building the final template can be expensive! Typically you only want to do this once.
    E.g.:

    ```python
    from string import Template
    from elasticai.creator.ir2verilog import type_handler, Code, TemplateDirector, Implementation

    class _ExplodingTemplate:
        def substitute(self, params):
            raise RuntimeError("Template not initialized!")

    _template = None

    @type_handler
    def fir_filter(impl: Implementation) -> Code:
        global _template
        if _template is None:
            _template = (
                TemplateDirector()
                .define_scoped_switch("USE_EXT_WEIGHTS", False)
                .define_scoped_switch("USE_EXT_MAC", False)
                .parameter("BITWIDTH")
                .parameter("LENGTH")
                .localparam("FILT_COEFFS")
                .localparam("NUM_MULT_UNIT")
                .add_module_name()
                .set_prototype(res.read_text(package_path, "verilog/filter_fir_full.v"))
                .build()
            )
        return _template.substitute(impl.attributes)
    ```
    """

    def __init__(self) -> None:
        self._builder = tpl.TemplateBuilder()
        self._define_switches: dict[str, bool] = {}
        self._module_name: dict[str, str] = {}
        self._delimiter = "§"

    def reset(self) -> Self:
        self._builder = tpl.TemplateBuilder()
        self._define_switches = {}
        self._module_name = {}
        return self

    def set_prototype(self, prototype: str) -> Self:
        self._builder.set_prototype(prototype)
        return self

    def parameter(self, name: str) -> Self:
        self._builder.add_parameter(_Parameter(name, self._delimiter))
        return self

    def localparam(self, name: str) -> Self:
        self._builder.add_parameter(_LocalParameter(name, self._delimiter))
        return self

    def replace_module_of_instance(self, module_name: str, new_name: str) -> Self:
        self._builder.add_parameter(_ModuleOfInstance(module_name, self._delimiter))
        self._module_name[module_name] = new_name
        return self

    def replace_instance_name(self, module_name: str, new_name: str) -> Self:
        self._builder.add_parameter(_InstanceName(module_name, self._delimiter))
        self._module_name[module_name] = new_name
        return self

    def define_scoped_switch(self, name: str, default: bool) -> Self:
        """Add a switch for a define that is scoped to the module name.

        The switch will be prefixed with the value that users provide as `module_name` to the render call.
        :param name:        String with name of the switch/define name
        :param default:     Setting switch for defining output state (True=set, False=undefine)
        """
        self._builder.add_parameter(_DefineSwitch(name, self._delimiter))
        self._builder.add_parameter(_IdParameter(name, self._delimiter))
        self._define_switches[name] = default
        return self

    def add_module_name(self) -> Self:
        self._builder.add_parameter(_ModuleName(self._delimiter))
        return self

    def build(self) -> "VerilogTemplate":
        MyStrTemplate = type(
            "MyStrTemplate", (_pyTemplate,), {"delimiter": self._delimiter}
        )
        return VerilogTemplate(
            MyStrTemplate(self._builder.build()),
            self._define_switches,
            self._module_name,
        )


class VerilogTemplate:
    def __init__(
        self,
        template: _pyTemplate,
        defines: dict[str, bool],
        module_name: dict[str, str],
    ) -> None:
        self._template = template
        self._defines = defines.copy()
        self._modules = module_name.copy()

    def substitute(self, params: Mapping[str, str | bool]) -> str:
        new_params: dict[str, str] = {}
        if "module_name" in params:
            module_prefix = f"{params['module_name']}_"
        else:
            module_prefix = ""

        for name, value in params.items():
            if name not in self._defines:
                new_params[name] = cast(str, value)

        defines: dict[str, bool] = self._defines.copy()
        for name, enabled in params.items():
            if name in self._defines and isinstance(enabled, bool):
                defines[name] = enabled

        modules: dict[str, str] = self._modules.copy()
        for module, instance in params.items():
            if module in self._modules and isinstance(instance, str):
                modules[module] = instance

        for name, enabled in defines.items():
            defname = f"def{name}"
            new_params[defname] = f"`define {module_prefix}{name}"
            new_params[name] = f"{module_prefix}{name}"
            if not enabled:
                new_params[defname] = "//" + cast(str, new_params[defname])

        for module, name in modules.items():
            new_params[module] = name

        return self._template.substitute(new_params)



from collections.abc import Callable
from typing import Iterable

from elasticai.creator import graph as g


class Graph:
    def __init__(self, wrapped: g.Graph[str], data: dict[str, str]):
        self.wrapped = wrapped
        self.data = data

    @property
    def nodes(self):
        return self.wrapped.nodes

    @property
    def edges(self):
        return list(self.wrapped.iter_edges())

    def as_dict(self):
        return {"nodes": set(self.nodes), "edges": set(self.edges)}


def build_graph_from_dict(
    d: dict[tuple[str, str], list[str]],
) -> Graph:
    graph_dict: dict[str, Iterable[str]] = {k[0]: v for k, v in d.items()}
    data_dict = {k[0]: k[1] for k in d}

    return Graph(g.BaseGraph.from_dict(graph_dict), data=data_dict)


def find_matches(graph, pattern) -> list[dict[str, str]]:
    def node_constraint(pattern_node, graph_node):
        return graph.data[graph_node] == pattern.data[pattern_node]

    return g.find_all_subgraphs(
        pattern=pattern.wrapped, graph=graph.wrapped, node_constraint=node_constraint
    )


class Matcher:
    def __init__(self, pattern: Graph):
        self.pattern = pattern
        self.graph: None | Graph = None

    def set_graph(self, graph: Graph):
        self.graph = graph

    def _node_constraint(self, pattern_node: str, graph_node: str) -> bool:
        assert self.graph is not None
        return self.graph.data[graph_node] == self.pattern.data[pattern_node]

    def __call__(self, graph: g.Graph[str], pattern: g.Graph[str]) -> dict[str, str]:
        return g.find_subgraph(
            pattern=pattern, graph=graph, node_constraint=self._node_constraint
        )


def get_rewriter(
    pattern: Graph,
    replacement: Graph,
    lhs: dict[str, str],
    rhs: dict[str, str],
) -> Callable[[Graph], Graph]:
    kwargs = locals()
    _rewrite = get_rewriter_returning_full_result(**kwargs)

    def rewrite(graph: Graph):
        return Graph(_rewrite(graph).new_graph, {})

    return rewrite


def get_rewriter_returning_full_result(
    pattern: Graph,
    replacement: Graph,
    lhs: dict[str, str],
    rhs: dict[str, str],
) -> Callable[[Graph], g.RewriteResult]:
    do_match = Matcher(pattern)

    def rewrite(graph: Graph) -> g.RewriteResult:
        do_match.set_graph(graph)
        match = do_match(graph.wrapped, pattern.wrapped)
        new_graph, new_names = g.rewrite(
            original=graph.wrapped,
            replacement=replacement.wrapped,
            match=match,
            lhs=lhs,
            rhs=rhs,
        )
        return g.RewriteResult(
            new_graph=new_graph, pattern_to_original=match, replacement_to_new=new_names
        )

    return rewrite



from .iterator_utils import batched


def structure_flat_CxN(vector, channels):
    """Assumes a vector of the following format
    (x[1, 1], ..., x[C, 1], x[1, 2], ..., x[C, 2], ..., x[1, N], ..., x[C, N])
    and reshapes it to a tensor y, such that
    y[i] contains all channels of step i
    """
    return tuple(batched(vector, channels))


def reshape_flat_CxN_groupwise(vector, channels, groups):
    """Assumes a vector of the following format
    (x[1, 1], ..., x[C, 1], x[1, 2], ..., x[C, 2], ..., x[1, N], ..., x[C, N])
    and reshapes it to a tensor y, such that
    y[i] contains a flat vector for all spatial steps and all channels in group i in the
    same CxN format.
    E.g.:
    >>> reshape_flat_CxN_groupwise(tuple(range(12)), channels=4, groups=2)
    ... ((0, 1, 4, 5, 8, 9), (2, 3, 6, 7, 10, 11))
    """
    structured = structure_flat_CxN(vector, channels)
    group_size = channels // groups
    grouped_spatial_steps = ((batched(x, group_size)) for x in structured)

    groups_combined_across_spatial_steps = zip(*grouped_spatial_steps)
    flattened_groups = tuple(
        tuple(y for x in xs for y in x) for xs in groups_combined_across_spatial_steps
    )
    return flattened_groups



from dataclasses import dataclass
from typing import cast, overload

from elasticai.creator.arithmetic.fxp_params import (
    ConvertableToFixedPointValues,
    FxpParams,
    T,
)


@dataclass
class FxpArithmetic:
    def __init__(self, fxp_params: FxpParams):
        self._config = fxp_params

    @property
    def config(self) -> FxpParams:
        return self._config

    @property
    def total_bits(self) -> int:
        return self._config.total_bits

    @property
    def frac_bits(self) -> int:
        return self._config.frac_bits

    @property
    def minimum_as_rational(self) -> float:
        return self._config.minimum_as_rational

    @property
    def maximum_as_rational(self) -> float:
        return self._config.maximum_as_rational

    def integer_out_of_bounds(self, number):
        return self._config.integer_out_of_bounds(number)

    @overload
    def cut_as_integer(self, number: float | int) -> int: ...

    @overload
    def cut_as_integer(self, number: list) -> list: ...

    @overload
    def cut_as_integer(self, number: T) -> T: ...

    def cut_as_integer(self, number: float | int | list | T) -> int | list | T:
        """Cutting input number to integer directly (more like in hardware)"""
        if isinstance(number, ConvertableToFixedPointValues):
            return self._cut_T_to_integer(cast(T, number))
        elif isinstance(number, list):
            return list(map(self.cut_as_integer, number))
        else:
            return self._cut_float_or_int_to_integer(number)

    def _cut_T_to_integer(self, number: T) -> T:
        return (number / self._config.minimum_step_as_rational).int().float()

    def _cut_float_or_int_to_integer(self, number: float | int) -> int:
        return int(number / self._config.minimum_step_as_rational)

    @overload
    def round_to_integer(self, number: float | int) -> int: ...

    @overload
    def round_to_integer(self, number: T) -> T: ...

    def round_to_integer(self, number: float | int | T) -> int | T:
        """Mathematical Round function for number"""
        if isinstance(number, ConvertableToFixedPointValues):
            return self._round_T_to_integer(cast(T, number))
        else:
            return self._round_float_or_int_to_integer(number)

    def _round_T_to_integer(self, number: T) -> T:
        return (number / self._config.minimum_step_as_rational).round().int().float()

    def _round_float_or_int_to_integer(self, number: float | int) -> int:
        return round(number / self._config.minimum_step_as_rational)

    @overload
    def as_rational(self, number: int) -> float: ...

    @overload
    def as_rational(self, number: T) -> T: ...

    def as_rational(self, number: int | T) -> float | T:
        return number * self._config.minimum_step_as_rational



from typing import cast

import pytest

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.identity.layer import BufferedIdentity
from elasticai.creator.nn.sequential.layer import Sequential
from elasticai.creator.vhdl.system_integrations.firmware_env5 import FirmwareENv5


def extract_skeleton_code(firmware) -> str:
    destination = InMemoryPath(name="build", parent=None)
    firmware.save_to(destination)
    srcs_dir = cast(InMemoryPath, destination["srcs"])
    actual_code_lines = cast(InMemoryFile, srcs_dir["skeleton"]).text
    actual_code = "\n".join(actual_code_lines)
    return actual_code


def test_firmware_generates_correct_skeleton_v1() -> None:
    model = Sequential(BufferedIdentity(num_input_features=10, total_bits=8))
    design = model.create_design("network")
    with pytest.warns(FutureWarning):
        firmware = FirmwareENv5(
            design, x_num_values=10, y_num_values=10, id=66, skeleton_version="v1"
        )

        actual_code = extract_skeleton_code(firmware)
    expected_code = """library ieee;
USE ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity skeleton is
    port (
        -- control interface
        clock                : in std_logic;
        clk_hadamard                : in std_logic;
        reset                : in std_logic; -- controls functionality (sleep)
        busy                : out std_logic; -- done with entire calculation
        wake_up             : out std_logic;
        -- indicate new data or request
        rd                    : in std_logic;    -- request a variable
        wr                 : in std_logic;     -- request changing a variable

        -- data interface
        data_in            : in std_logic_vector(7 downto 0);
        address_in        : in std_logic_vector(15 downto 0);
        data_out            : out std_logic_vector(7 downto 0);

        debug                : out std_logic_vector(7 downto 0);

        led_ctrl             : out std_logic_vector(3 DOWNTO 0)
    );
end;

architecture rtl of skeleton is
    constant DATA_WIDTH_IN : integer := 8;
    constant DATA_WIDTH_OUT : integer := 8;
    constant X_ADDR_WIDTH : integer := 4;
    constant X_NUM_VALUES : integer := 10;
    constant Y_ADDR_WIDTH : integer:= 4;
    constant Y_NUM_VALUES : integer := 10;

    signal network_enable :  std_logic;

    signal c_config_en :  std_logic;
    signal done :  std_logic;

    signal x :  std_logic_vector(DATA_WIDTH_IN-1 downto 0);
    signal y : std_logic_vector(DATA_WIDTH_OUT-1 downto 0);
    signal x_address :  std_logic_vector(X_ADDR_WIDTH-1 downto 0);
    signal y_address :  std_logic_vector(Y_ADDR_WIDTH-1 downto 0);

    type buf_data_in_t is array (0 to X_NUM_VALUES) of std_logic_vector(DATA_WIDTH_IN-1 downto 0);
    signal data_buf_in : buf_data_in_t;
    type skeleton_id_data_t is array (0 to 0) of std_logic_vector(7 downto 0);
    signal skeleton_id_str : skeleton_id_data_t := (0 => "01000010");

    function pad_output_to_middleware(network_out : std_logic_vector(DATA_WIDTH_OUT-1 downto 0)) return std_logic_vector is
    variable k : std_logic_vector(7 downto 0);
    begin
        if DATA_WIDTH_OUT /= 8 then
            k(7 downto DATA_WIDTH_OUT) := (others => '0');
        end if;
        k(DATA_WIDTH_OUT-1 downto 0) := network_out;
        return k;
    end function;

begin

    i_network: entity work.network(rtl)
    port map (
        clock => clock,
        enable => network_enable,
        x => x,
        x_address => x_address,
        y => y,
        y_address => y_address,
        done => done
    );

    -- orignial implementation
    busy <= not done;
    wake_up <= done;

    receive_data_from_middleware: process (clock, wr, address_in)
    variable int_addr : integer range 0 to 20000;
    begin
        if rising_edge(clock) then
            if reset = '1' then
                network_enable <= '0';
            else
                int_addr := to_integer(unsigned(address_in));
                if int_addr < X_NUM_VALUES then
                    data_buf_in(int_addr) <= data_in(DATA_WIDTH_IN-1 downto 0);
                elsif int_addr = 100 then
                    network_enable <= data_in(0);
                end if;
            end if;
        end if;
    end process;

    sendback_data_to_middleware: process  (clock, rd, address_in)
    variable int_addr : integer range 0 to 20000;
    begin
        if rising_edge(clock) then
            int_addr := to_integer(unsigned(address_in));
            if int_addr <= Y_NUM_VALUES-1 then
                y_address <= address_in(y_address'length-1 downto 0);
                data_out(7 downto 0) <= pad_output_to_middleware(y);
            elsif int_addr = 2000  then
                data_out(7 downto 0) <= skeleton_id_str(int_addr-2000);
            end if;
        end if;
    end process;

    send_buf_to_network: process (clock, x_address)
    begin
        if rising_edge(clock) then
            x <= data_buf_in(to_integer(unsigned(x_address)));
        end if;
    end process;

end rtl;"""

    assert expected_code == actual_code


def test_firmware_generates_correct_skeleton_v2() -> None:
    model = Sequential(BufferedIdentity(num_input_features=10, total_bits=8))
    design = model.create_design("network")
    firmware = FirmwareENv5(
        design,
        x_num_values=10,
        y_num_values=10,
        id=list(range(0, 16)),
        skeleton_version="v2",
    )
    actual_code = extract_skeleton_code(firmware)
    expected_code = """library ieee;
USE ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity skeleton is
    port (
        -- control interface
        clock                : in std_logic;
        clk_hadamard                : in std_logic;
        reset                : in std_logic; -- controls functionality (sleep)
        busy                : out std_logic; -- done with entire calculation
        wake_up             : out std_logic;
        -- indicate new data or request
        rd                    : in std_logic;    -- request a variable
        wr                 : in std_logic;     -- request changing a variable

        -- data interface
        data_in            : in std_logic_vector(7 downto 0);
        address_in        : in std_logic_vector(15 downto 0);
        data_out            : out std_logic_vector(7 downto 0);

        debug                : out std_logic_vector(7 downto 0);

        led_ctrl             : out std_logic_vector(3 DOWNTO 0)
    );
end;

architecture rtl of skeleton is
    constant DATA_WIDTH_IN : integer := 8;
    constant DATA_WIDTH_OUT : integer := 8;
    constant X_ADDR_WIDTH : integer := 4;
    constant X_NUM_VALUES : integer := 10;
    constant Y_ADDR_WIDTH : integer:= 4;
    constant Y_NUM_VALUES : integer := 10;

    signal network_enable :  std_logic;

    signal c_config_en :  std_logic;
    signal done :  std_logic;

    signal x :  std_logic_vector(DATA_WIDTH_IN-1 downto 0);
    signal y : std_logic_vector(DATA_WIDTH_OUT-1 downto 0);
    signal x_address :  std_logic_vector(X_ADDR_WIDTH-1 downto 0);
    signal y_address :  std_logic_vector(Y_ADDR_WIDTH-1 downto 0);

    type buf_data_in_t is array (0 to X_NUM_VALUES) of std_logic_vector(DATA_WIDTH_IN-1 downto 0);
    signal data_buf_in : buf_data_in_t;
    type skeleton_id_data_t is array (0 to 15) of std_logic_vector(7 downto 0);
    signal skeleton_id_str : skeleton_id_data_t := ("00000000", "00000001", "00000010", "00000011", "00000100", "00000101", "00000110", "00000111", "00001000", "00001001", "00001010", "00001011", "00001100", "00001101", "00001110", "00001111");

    function pad_output_to_middleware(network_out : std_logic_vector(DATA_WIDTH_OUT-1 downto 0)) return std_logic_vector is
    variable k : std_logic_vector(7 downto 0);
    begin
        if DATA_WIDTH_OUT /= 8 then
            k(7 downto DATA_WIDTH_OUT) := (others => '0');
        end if;
        k(DATA_WIDTH_OUT-1 downto 0) := network_out;
        return k;
    end function;

begin

    i_network: entity work.network(rtl)
    port map (
        clock => clock,
        enable => network_enable,
        x => x,
        x_address => x_address,
        y => y,
        y_address => y_address,
        done => done
    );

    busy <= not done;
    wake_up <= done;

    receive_data_from_middleware: process (clock)
    variable int_addr : integer range 0 to 18 + X_NUM_VALUES;
    begin
        if rising_edge(clock) then
            if reset = '1' then
                network_enable <= '0';
            else
                if wr = '1' then
                    int_addr := to_integer(unsigned(address_in));
                    if int_addr = 16 then
                        network_enable <= data_in(0);
                    elsif int_addr >= 18 and int_addr < 18 + X_NUM_VALUES then
                        data_buf_in(int_addr-18) <= data_in(DATA_WIDTH_IN-1 downto 0);
                    end if;
                end if;
            end if;
        end if;
    end process;

    sendback_data_to_middleware: process  (clock)
    variable int_addr : integer range 0 to 18 + Y_NUM_VALUES;
    begin
        if rising_edge(clock) then
            if rd = '1' then
                int_addr := to_integer(unsigned(address_in));
                if int_addr <= 15 then
                    data_out(7 downto 0) <= skeleton_id_str(int_addr);
                elsif int_addr >= 18 and int_addr < 18 + Y_NUM_VALUES then
                    y_address <= std_logic_vector(unsigned(address_in(y_address'length-1 downto 0))-18);
                    data_out(7 downto 0) <= pad_output_to_middleware(y);
                end if;
            end if;
        end if;
    end process;

    send_buf_to_network: process (clock)
    begin
        if falling_edge(clock) then
            x <= data_buf_in(to_integer(unsigned(x_address)));
        end if;
    end process;

end rtl;"""
    assert expected_code == actual_code
